static char rcsid[] = "$Id$";
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>		/* For isdigit */
#include <math.h>		/* For fabs and sqrt */

#include "assert.h"
#include "bool.h"
#include "types.h"
#include "iit-read.h"
#include "transcriptome.h"
#include "orderstat.h"

#include "getline.h"
#include "samread.h"
#include "bitvector.h"
#include "intlist.h"
#include "uintlist.h"
#include "list.h"
#include "doublelist.h"
#include "table.h"
#include "tableint.h"
#include "tableuint.h"
#include "uinttableuint.h"
#include "uinttable.h"
#include "equiv-class-bulk.h"

#include "datadir.h"
#include "getopt.h"

#define TWO_STEP_EM 1

#define nottab(x) (x) != '\t'

#ifdef DEBUG
#define debug(x) x
#else
#define debug(x)
#endif



/* Need genomedir to get transcript_map_iit, which has the gene symbols */
static char *user_genomedir = NULL;
static char *genome_dbroot = NULL;
static char *user_transcriptomedir = NULL;
static char *transcriptome_dbroot = NULL;

static char *input_command = NULL;
static char *output_prefix = NULL;
static bool nonzero_only_p = false;


static struct option long_options[] = {
  /* Input options */
  {"dir", required_argument, 0, 'D'},	/* user_genomedir */
  {"db", required_argument, 0, 'd'}, /* genome_dbroot */
  {"transcriptdir", required_argument, 0, 'C'},	/* user_transcriptomedir */
  {"transcriptdb", required_argument, 0, 'c'}, /* transcriptome_dbroot */
  {"input-command", required_argument, 0, 0},		       /* input_command */

  /* Output options */
  {"output", required_argument, 0, 'o'}, /* output_prefix */
  {"nonzero-only", no_argument, 0, 0},	 /* nonzero_only_p */

  {"version", no_argument, 0, 0}, /* print_program_version */
  {"help", no_argument, 0, 0}, /* print_program_usage */
  {0, 0, 0, 0}
};

static void
print_program_version () {
  char *genomedir;

  fprintf(stdout,"\n");
  fprintf(stdout,"bulk-counts: Parses GSNAP output from transcriptome-guided genomic alignment\n");
  fprintf(stdout,"Part of GMAP package, version %s\n",PACKAGE_VERSION);

  fprintf(stdout,"Default gmap directory (compiled): %s\n",GMAPDB);
  genomedir = Datadir_find_genomedir(/*user_genomedir*/NULL);
  fprintf(stdout,"Default gmap directory (environment): %s\n",genomedir);
  FREE(genomedir);

  fprintf(stdout,"Thomas D. Wu, Genentech, Inc.\n");
  fprintf(stdout,"Contact: twu@gene.com\n");
  fprintf(stdout,"\n");

  return;
}

static void
print_program_usage ();


/* transcript_dict: transcript_name -> transcript_index */
static Tableuint_T
get_transcripts (unsigned int *n_total_transcripts, IIT_T transcript_map_iit) {
  Tableuint_T transcript_dict;
  char *transcript_name, *transcript_name_ptr;
  unsigned int indx;
  bool allocp;

  *n_total_transcripts = (unsigned int) IIT_total_nintervals(transcript_map_iit);

  transcript_dict = Tableuint_new(/*hint*/(int) *n_total_transcripts,
                                  Table_string_compare,Table_string_hash);

  for (indx = 1; indx <= *n_total_transcripts; indx++) {
    transcript_name_ptr = IIT_label(transcript_map_iit,indx,&allocp);
    transcript_name = (char *) MALLOC((strlen(transcript_name_ptr)+1)*sizeof(char));
    strcpy(transcript_name,transcript_name_ptr);
    Tableuint_put(transcript_dict,(void *) transcript_name,indx); /* Store as 1-based */
  }

  return transcript_dict;
}


static int *
get_exon_info (int **transcript_nexons, int ***exonbounds, Chrpos_T ***exonstarts,
	       Transcriptome_T transcriptome, unsigned int n_total_transcripts) {
  int *lengths;
  unsigned int i;
  int nexons;

  lengths = (int *) MALLOC(n_total_transcripts*sizeof(int));
  *transcript_nexons = (int *) MALLOC(n_total_transcripts*sizeof(int));
  *exonbounds = (int **) MALLOC(n_total_transcripts*sizeof(int *));
  *exonstarts = (Chrpos_T **) MALLOC(n_total_transcripts*sizeof(Chrpos_T *));
  for (i = 0; i < n_total_transcripts; i++) {
    (*transcript_nexons)[i] = nexons =
      Transcriptome_exons(&((*exonbounds)[i]),&((*exonstarts)[i]),transcriptome,/*trnum*/i+1);
    lengths[i] = (*exonbounds)[i][nexons - 1];
  }

  return lengths;
}



static char *
transcript_gene_name (int *stringlen, char *annot) {
  char *gene, *p;

  /* Parse header */
  p = annot;
  while (*p != '\0' && *p != '\n' && !isspace(*p)) {
    p++;
  }
  *stringlen = p - annot;

  gene = (char *) MALLOC(((*stringlen)+1)*sizeof(char));
  strncpy(gene,annot,*stringlen);
  gene[*stringlen] = '\0';

  return gene;
}


/* transcripti_to_genei: transcript_index -> gene_index */
static char **
get_genes (unsigned int *n_total_genes, unsigned int **transcripti_to_genei,
	   Uintlist_T **genei_to_transcripti, IIT_T transcript_map_iit) {
  char **all_genes;
  List_T gene_list = NULL;
  Uintlist_T index_list = NULL;

  unsigned int gene_index, ngenes = 0, n_total_transcripts,
    transcript_index, indx;
  int gene_namelen;
  int ignore;
  Tableuint_T gene_table;
  char *annot, *restofheader, *symbol;
  bool allocp;

  n_total_transcripts = (unsigned int) IIT_total_nintervals(transcript_map_iit);
  gene_table = Tableuint_new(/*hint*/(int) n_total_transcripts,
			     Table_string_compare,Table_string_hash);

  for (indx = 1; indx <= n_total_transcripts; indx++) {
    annot = IIT_annotation(&restofheader,transcript_map_iit,indx,&allocp);
    symbol = (char *) transcript_gene_name(&gene_namelen,annot);
    /* gene_table stores gene-indices as 1-based, so 0 means not found */
    if ((gene_index = Tableuint_get(gene_table,(void *) symbol)) == 0) {
      Tableuint_put(gene_table,(void *) symbol,++ngenes);
      gene_list = List_push(gene_list,(void *) symbol);
      index_list = Uintlist_push(index_list,ngenes-1); /* 0-based */
    } else {
      index_list = Uintlist_push(index_list,gene_index-1); /* 0-based */
      FREE(symbol);
    }
    if (allocp) {
      FREE(restofheader);
    }
  }

  *n_total_genes = ngenes;
  gene_list = List_reverse(gene_list);
  all_genes = (char **) List_to_array(gene_list,NULL);
  List_free(&gene_list);

  index_list = Uintlist_reverse(index_list);
  *transcripti_to_genei = Uintlist_to_array(&ignore,index_list); /* 0-based */
  Uintlist_free(&index_list);

  *genei_to_transcripti = (Uintlist_T *) CALLOC(ngenes,sizeof(Uintlist_T));
  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    gene_index = (*transcripti_to_genei)[transcript_index]; /* 0-based */
    index_list = (*genei_to_transcripti)[gene_index];
    (*genei_to_transcripti)[gene_index] = Uintlist_push(index_list,transcript_index); /* 0-based */
  }
  
  Tableuint_free(&gene_table);

  return all_genes;
}


/* Format for string: I3-4: */
/* assoc_list is a list of Assoc_T objects */
static List_T
add_introns_from_string (Assoc_T *assoc, List_T assoc_list, int nexons, char *string) {
  char *p = string;
  unsigned int introni;
  Bitvector_T introns;

  introns = Bitvector_new(/*nintrons*/nexons - 1);

  while (*p != ':' && *p != '/' && sscanf(p,"%u",&introni) > 0) {
    Bitvector_set(introns,introni-1); /* 0-based */

    while (*p != ':' && *p != '/' && *p != '-') {
      p++;
    }
    if (*p == '-') {
      p++;
    }
  }

  if ((*assoc = Assoc_find(introns,assoc_list)) != NULL) {
    Bitvector_free(&introns);
  } else {
    *assoc = Assoc_new(introns,/*value*/(Intlist_T) NULL);
    assoc_list = List_push(assoc_list,(void *) *assoc);
  }

  return assoc_list;
}


static void
lines_gc (List_T *lines) {
  List_T p;
  char *line;

  for (p = *lines; p != NULL; p = List_next(p)) {
    line = (char *) List_head(p);
    FREE(line);
  }
  List_free(&(*lines));
  *lines = (List_T) NULL;

  return;
}


/* This needs to allocate the string */
static char *
get_acc (char *line) {
  char *acc, *p;
  int length;

  p = line;
  while (nottab(*p++)) ;
  length = (p - 1 - line)/sizeof(char);
  acc = (char *) MALLOC((length+1)*sizeof(char));
  strncpy(acc,line,length);
  acc[length] = '\0';

  return acc;
}


/* Approach: Read alignment file two times: (1) to find the mean and
   standard deviation of fragment lengths, and to find find retained
   introns, with fragment length support for them.  We then filter the
   retained introns.  Then (2) again to assign velocities as spliced
   across those introns, unspliced, or ambiguous with respect to those
   introns */
/* For single-end reads, need to search both XX and XY fields */
/* For paired-end reads, search XY field only if XX is NULL, which means primary support for introns */

/************************************************************************
 *  all_transcript_introns is a List_T *, each elt being a list of Bitvector_T
 *  It is used as an "x", with the "y" being a Doublelist_T * (lengths, weights, alpha)
 * 
 *  Intermediate structure is all_transcript_introns_support, which is a List_T *,
 *  each elt being a list of Assoc_T, with a Bitvector_T and an Intlist_T
 ************************************************************************/

static void
pre_process (bool *identicalp, bool *splicedp, bool *unsplicedp, int *prev_fragment_length,
	     unsigned int flag, int fragment_length, char *transcript_list,
	     List_T *all_transcript_introns_support, int *transcript_nexons, Tableuint_T transcript_dict) {
  char velocity_char, *p, *q;
  int trlength;

  char *transcript_name;
  int nexons;
  List_T assocs;
  unsigned int transcript_index;
  int namelen;
  Assoc_T assoc;


  p = transcript_list;
  
  while (*p != '\0') {
    /* Format: +ENST00000456328.2:270..369:S:.1s|s2.:264 */
    p++;			/* Skip [+-] */
    q = p;
    while (*q++ != ':') {
      /* Advance through transcript name */
    }
    namelen = q - p - 1;
    while (*q++ != ':') {
      /* Advance through transcript coordinates */
    }

    if ((velocity_char = *q) == 'I') {
      transcript_name = (char *) MALLOC((namelen+1)*sizeof(char));
      strncpy(transcript_name,p,namelen);
      transcript_name[namelen] = '\0';
      transcript_index = Tableuint_get(transcript_dict,(void *) transcript_name) - 1; /* 0-based */
      
      /* Revise introns globally in all_transcript_introns_support */
      assocs = all_transcript_introns_support[transcript_index];
      nexons = transcript_nexons[transcript_index];
      all_transcript_introns_support[transcript_index] =
	add_introns_from_string(&assoc,assocs,nexons,/*string*/&(q[1]));
      
      /* Store fragment length (genomic, not transcriptomic), and must be a paired-end, concordant read */
      if ((flag & PAIRED_MAPPING) && fragment_length > 0) {
	assoc->value = Intlist_push(assoc->value,fragment_length);
      }
      
      FREE(transcript_name);
      *unsplicedp = true;

    } else if (velocity_char == 'S') {
      *splicedp = true;
    }
    
    if (velocity_char == 'X') {
      /* Ignore, since trlength is not valid */
    } else if ((flag & PAIRED_READ) == 0) {
      /* No trlengths printed */
      *prev_fragment_length = fragment_length; /* readlength, which will be the same on all lines */
    } else if ((flag & PAIRED_MAPPING) == 0) {
      /* No trlengths printed */
    } else {
      while (*q++ != ':') {
	/* Advance through velocity information */
      }
      while (*q++ != ':') {
	/* Advance through gene structure */
      }
      if (sscanf(q,"%d",&trlength) == 0) {
	fprintf(stderr,"%s",transcript_list);
	abort();
      } else if (*prev_fragment_length == 0) {
	*prev_fragment_length = trlength;
      } else if (trlength != *prev_fragment_length) {
	*identicalp = false;
      }
    }
    
    /* Advance to next transcript */
    p = q;
    while (*p != '\0' && *p != ',') {
      p++;
    }
    if (*p == ',') {
      p++;
    }
  }

  return;
}


/* Process only for fragment_length.  Must be a paired-end read. */
static void
pre_process_for_fraglen (bool *identicalp, bool *unsplicedp, int *prev_fragment_length,
			 unsigned int flag, int fragment_length, char *transcript_list) {
  char velocity_char, *p, *q;
  int trlength;

  p = transcript_list;

  while (*p != '\0') {
    /* Format: +ENST00000456328.2:270..369:S:.1s|s2.:264 */
    p++;			/* Skip [+-] */
    q = p;
    while (*q++ != ':') {
      /* Advance through transcript name */
    }
    /* namelen = q - p - 1; */
    while (*q++ != ':') {
      /* Advance through transcript coordinates */
    }
    
    if ((velocity_char = *q) == 'I') {
      *unsplicedp = true;
    }
    
    if (*q == 'X') {
      /* Ignore, since trlength is not valid */
    } else if ((flag & PAIRED_READ) == 0) {
      /* No trlengths printed */
      *prev_fragment_length = fragment_length; /* readlength, which will be the same on all lines */
    } else if ((flag & PAIRED_MAPPING) == 0) {
      /* No trlengths printed */
    } else {
      while (*q++ != ':') {
	/* Advance through velocity information */
      }
      while (*q++ != ':') {
	/* Advance through gene structure */
      }
      if (sscanf(q,"%d",&trlength) == 0) {
	fprintf(stderr,"%s",transcript_list);
	abort();
      } else if (*prev_fragment_length == 0) {
	*prev_fragment_length = trlength;
      } else if (trlength != *prev_fragment_length) {
	*identicalp = false;
      }
    }
    
    /* Advance to next transcript */
    p = q;
    while (*p != '\0' && *p != ',') {
      p++;
    }
    if (*p == ',') {
      p++;
    }
  }

  return;
}


/* consistent_spliced_p solves the issue where a transcript has an exon that subsumes an intron */
static Intlist_T
pass1_block (int *nfragments, double *mean, double *M2, Intlist_T fragment_lengths,
	     List_T *all_transcript_introns_support, List_T lines,
	     int *transcript_nexons, Tableuint_T transcript_dict) {
  bool identicalp = true, splicedp = false, unsplicedp = false, consistent_spliced_p;
  int prev_fragment_length = 0, fragment_length;
  double delta, delta2;

  List_T l;
  char *line, *auxinfo;
  char *transcripts_consistent, *transcripts_inconsistent;
  unsigned int flag;


  /* Get fragment length and retained introns */
  for (l = lines; l != NULL; l = List_next(l)) {
    line = (char *) List_head(l);
    auxinfo = Samread_auxinfo(&flag,&fragment_length,line);

    if ((transcripts_consistent = Samread_get_aux_string("XX",auxinfo)) != NULL) {
      pre_process(&identicalp,&splicedp,&unsplicedp,&prev_fragment_length,
		  flag,fragment_length,transcripts_consistent,all_transcript_introns_support,
		  transcript_nexons,transcript_dict);
      FREE(transcripts_consistent);
    }
  }

  consistent_spliced_p = splicedp;

  for (l = lines; l != NULL; l = List_next(l)) {
    line = (char *) List_head(l);
    auxinfo = Samread_auxinfo(&flag,&fragment_length,line);

    if ((transcripts_inconsistent = Samread_get_aux_string("XY",auxinfo)) == NULL) {
      /* Skip */

    } else if ((flag & PAIRED_MAPPING) == 0) {
      /* Always process for single-end */
      pre_process(&identicalp,&splicedp,&unsplicedp,&prev_fragment_length,
		  flag,fragment_length,transcripts_inconsistent,all_transcript_introns_support,
		  transcript_nexons,transcript_dict);
      FREE(transcripts_inconsistent);

    } else if (consistent_spliced_p == false) {
      pre_process(&identicalp,&splicedp,&unsplicedp,&prev_fragment_length,
		  flag,fragment_length,transcripts_inconsistent,all_transcript_introns_support,
		  transcript_nexons,transcript_dict);
      FREE(transcripts_inconsistent);

    } else {
      pre_process_for_fraglen(&identicalp,&unsplicedp,&prev_fragment_length,
			      flag,fragment_length,transcripts_inconsistent);
      FREE(transcripts_inconsistent);
    }
  }

  /* Welford's algorithm */
  if (identicalp == true && unsplicedp == false && prev_fragment_length != 0) {
    fragment_lengths = Intlist_push(fragment_lengths,prev_fragment_length);
    /* printf("%d %s",prev_fragment_length,line); */
    *nfragments += 1;
    if (*nfragments == 1) {
      *mean = prev_fragment_length;
      *M2 = 0.0;
    } else {
      delta = prev_fragment_length - (*mean);
      *mean += delta / (double) *nfragments;
      delta2 = prev_fragment_length - (*mean);
      *M2 += delta * delta2;
    }
  }

  return fragment_lengths;
}


static bool
intron_supported_p (Intlist_T support_lengths, double median_fragment_length,
		    double robust_sd_fragment_length) {
  Intlist_T p;
  double dev;

  for (p = support_lengths; p != NULL; p = Intlist_next(p)) {
    dev = ((double) Intlist_head(p) - median_fragment_length) / robust_sd_fragment_length;
    if (dev > -3.0 && dev < 3.0) {
      return true;
    }
  }

  return false;
}


static List_T *
pass1 (double *median_fragment_length, double *robust_sd_fragment_length,
       double *mean_fragment_length, double *sd_fragment_length, FILE *fp,
       int *transcript_nexons, Tableuint_T transcript_dict, char **all_transcripts,
       unsigned int n_total_transcripts) {
  List_T *all_transcript_introns; /* List_T * of Bitvector_T objects */
  List_T *all_transcript_introns_support, p; /* List_T * of Assoc_T objects, with value being an Intlist_T */
  int nfragments = 0, n, i;
  unsigned int transcript_index;

  Assoc_T assoc;
  Intlist_T intron_support_lengths;
  double mean, M2, mad;
  int median;
  Intlist_T fragment_lengths = NULL;
  int *array;

  List_T lines;
  char *line, *acc, *lastacc;
  int linelen;

  all_transcript_introns_support = (List_T *) CALLOC(n_total_transcripts,sizeof(List_T));
  
  lastacc = (char *) MALLOC(1*sizeof(char));
  lastacc[0] = '\0';
  lines = (List_T) NULL;
  while ((line = Getline_wlinefeed(&linelen,fp)) != NULL) {
    acc = get_acc(line);
    if (!strcmp(acc,lastacc)) {
      /* Same acc as before */
      FREE(acc);
      acc = lastacc;
    } else if (lines == NULL) {
      /* First acc seen */
      FREE(lastacc);
      lastacc = acc;
    } else {
      /* Process lines */
      fragment_lengths = pass1_block(&nfragments,&mean,&M2,fragment_lengths,
				     all_transcript_introns_support,lines,transcript_nexons,
				     transcript_dict);
      lines_gc(&lines);
      FREE(lastacc);
      lastacc = acc;
    }
    lines = List_push(lines,(void *) line);
  }

  if (lines == NULL) {
    /* No lines seen */
    FREE(lastacc);
  } else {
    /* Process lines */
    fragment_lengths = pass1_block(&nfragments,&mean,&M2,fragment_lengths,
				   all_transcript_introns_support,lines,transcript_nexons,
				   transcript_dict);
    lines_gc(&lines);
    FREE(lastacc);
  }

  array = Intlist_to_array(&n,fragment_lengths);
  Intlist_free(&fragment_lengths);

  median = Orderstat_int_pct(array,n,/*percentile*/0.50);
  *median_fragment_length = (double) median;
  for (i = 0; i < n; i++) {
    if (array[i] > median) {
      array[i] = array[i] - median;
    } else {
      array[i] = median - array[i];
    }
  }
  mad = Orderstat_int_pct(array,n,/*percentile*/0.50);
  *robust_sd_fragment_length = 1.4826 * (double) mad;
  FREE(array);

  if (nfragments == 0) {
    *mean_fragment_length = 0.0;
    *sd_fragment_length = 0.0;
  } else {
    *mean_fragment_length = mean;
    *sd_fragment_length = sqrt(M2/nfragments);
  }


  /* Use fragment length information to filter Bitvector_T objects in all_transcript_introns_support */
  all_transcript_introns = (List_T *) CALLOC(n_total_transcripts,sizeof(List_T));

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    for (p = all_transcript_introns_support[transcript_index]; p != NULL; p = List_next(p)) {
      assoc = (Assoc_T) List_head(p);

#ifdef DEBUG
      printf("%u ",transcript_index);
      Bitvector_print_file(stdout,assoc->bitvector,/*sep*/',');
      printf("\n");
#endif

      if ((intron_support_lengths = (Intlist_T) assoc->value) == NULL) {
	/* Must have had only single-end support, so accept */
	all_transcript_introns[transcript_index] = List_push(all_transcript_introns[transcript_index],(void *) assoc->bitvector);

      } else if (intron_supported_p(intron_support_lengths,*median_fragment_length,
				    *robust_sd_fragment_length) == false) {
	fprintf(stderr,"Removing intron support for %s\n",all_transcripts[transcript_index]);
	Intlist_free(&intron_support_lengths);
	Bitvector_free(&assoc->bitvector);

      } else {
	all_transcript_introns[transcript_index] = List_push(all_transcript_introns[transcript_index],(void *) assoc->bitvector);
	Intlist_free(&intron_support_lengths);
      }

      Assoc_free(&assoc);
    }
    List_free(&(all_transcript_introns_support[transcript_index]));

    /* Should we sort the bitvectors? */
  }

  FREE(all_transcript_introns_support);

  return all_transcript_introns;
}


#if 0
/* string points to velocity_char */
/* exons and spliced_introns are 1-based */
static Intlist_T
read_tokens (Intlist_T *spliced_introns, char *string, char transcript_sensedir) {
  Intlist_T exons = NULL;
  char *p = string, *q, firstchar, lastchar;
  int exoni;

  *spliced_introns = (Intlist_T) NULL;

  while (*p++ != ':') {
    /* Advance through velocity_char */
  }

  while (*p != '\0' && *p != ',') {
    q = p;
    while (*q != '\0' && *q != ',' && *q != '|') {
      q++;
    }

    if (sscanf(&(p[1]),"%d",&exoni) > 0) {
      exons = Intlist_push(exons,exoni);

      firstchar = p[0];
      lastchar = q[-1];
      if (firstchar == 's') {
	if (transcript_sensedir == '+') {
	  *spliced_introns = Intlist_push(*spliced_introns,exoni-1);
	} else if (transcript_sensedir == '-') {
	  *spliced_introns = Intlist_push(*spliced_introns,exoni);
	} else {
	  abort();
	}
      }
      if (lastchar == 's') {
	if (transcript_sensedir == '+') {
	  *spliced_introns = Intlist_push(*spliced_introns,exoni);
	} else if (transcript_sensedir == '-') {
	  *spliced_introns = Intlist_push(*spliced_introns,exoni-1);
	} else {
	  abort();
	}
      }
    }

    if (*q == '|') {
      q++;
    }
    p = q;
  }

  *spliced_introns = Intlist_reverse(*spliced_introns);
  return Intlist_reverse(exons);
}
#endif


/* Format for string: IS/S3-4: */
static void
introns_from_string (Bitvector_T *retained_introns, Bitvector_T *spliced_introns,
		     int nexons, char *string) {
  char *p = string, velocity_char;
  unsigned int x;

  *retained_introns = *spliced_introns = (Bitvector_T) NULL;

  while (*p != ':') {
    if ((velocity_char = *p) == 'X' || velocity_char == 'U') {
      /* GSNAP output should have put X or U before I or S */
      abort();

    } else if (velocity_char == 'A') {
      return;

    } else if (velocity_char == 'I') {
      p++;
      while (*p != ':' && *p != '/' && sscanf(p,"%u",&x) > 0) {
	if (*retained_introns == (Bitvector_T) NULL) {
	  *retained_introns = Bitvector_new(/*nintrons*/nexons - 1);
	}
	Bitvector_set(*retained_introns,x-1); /*0-based*/

	while (*p != ':' && *p != '/' && *p != '-') {
	  p++;
	}
	if (*p == '-') {
	  p++;
	}
      }

    } else if (velocity_char == 'S') {
      p++;
      while (*p != ':' && *p != '/' && sscanf(p,"%u",&x) > 0) {
	if (*spliced_introns == (Bitvector_T) NULL) {
	  *spliced_introns = Bitvector_new(/*nintrons*/nexons - 1);
	}
	Bitvector_set(*spliced_introns,x-1); /*0-based*/

	while (*p != ':' && *p != '/' && *p != '-') {
	  p++;
	}
	if (*p == '-') {
	  p++;
	}
      }

    } else {
      fprintf(stderr,"Unrecognized velocity char %c in %s\n",velocity_char,string);
      abort();
    }
      
    if (*p == '/') {
      p++;
    }
  }

  return;
}


static void
process_transcripts (bool *block_splicedp,
		     Uinttableuint_T transcript_status, Uinttable_T retained_introns_table,
		     unsigned int flag, char *transcript_list,
		     int *transcript_nexons, Tableuint_T transcript_dict, List_T *all_transcript_introns,
		     double median_fragment_length, double robust_sd_fragment_length) {

  char *transcript_name, main_velocity_char;
  Bitvector_T retained_introns, spliced_introns;
  unsigned int transcript_index;
  int nexons;
  char *p, *q;
  int namelen;

  int trlength;
  double dev;


  p = transcript_list;
  while (*p != '\0') {
    p++; /* Skip [+-] */
    q = p;
    while (*q++ != ':') {
      /* Advance through transcript name */
    }
    namelen = q - p - 1;
    while (*q++ != ':') {
      /* Advance through transcript coordinates */
    }

    if ((main_velocity_char = *q) == 'X' || main_velocity_char == 'U') {
      /* Skip */

    } else {
      transcript_name = (char *) MALLOC((namelen+1)*sizeof(char));
      strncpy(transcript_name,p,namelen);
      transcript_name[namelen] = '\0';
      transcript_index = Tableuint_get(transcript_dict,(void *) transcript_name) - 1; /* 0-based */
      nexons = transcript_nexons[transcript_index];

      if (Uinttableuint_get(transcript_status,transcript_index) != UNSEEN) {
	/* Already seen, and GSNAP replicates introns onto both ends */
	debug(printf("Already saw status for transcript index %u\n",transcript_index));

      } else {
	trlength = 0;
	dev = 0.0;
	p = q;
	
	if ((flag & PAIRED_READ) == 0) {
	  /* No trlengths printed */
	} else if ((flag & PAIRED_MAPPING) == 0) {
	  /* No trlengths printed */
	} else if (robust_sd_fragment_length == 0) {
	  /* Cannot compute dev */
	} else {
	  while (*q++ != ':') {
	    /* Advance through velocity information */
	  }
	  while (*q++ != ':') {
	    /* Advance through gene structure */
	  }
	  if (sscanf(q,"%d",&trlength) == 0) {
	    fprintf(stderr,"%s",transcript_list);
	    abort();
	  } else {
	    dev = ((double) trlength - median_fragment_length) / robust_sd_fragment_length;
	  }
	}
	/* Cannot use q anymore */
      
	/* Determine whether there is splicing relative to the retained intron(s) */
	if (dev >= 1000.0) {
	  /* Ignore based on the fragment length */
	  
	} else if (main_velocity_char == 'S') {
	  Uinttableuint_put(transcript_status,transcript_index,SPLICED);
	  Uinttable_put(retained_introns_table,transcript_index,(void *) NULL);
	  debug(printf(" => SPLICED %d %s\n",transcript_index,transcript_name));
	  *block_splicedp = true;

	} else if (main_velocity_char == 'A') {
	  Uinttableuint_put(transcript_status,transcript_index,BOTH);
	  Uinttable_put(retained_introns_table,transcript_index,(void *) NULL);
	  debug(printf(" => BOTH %d %s\n",transcript_index,transcript_name));

	} else if (main_velocity_char == 'I') {
	  introns_from_string(&retained_introns,&spliced_introns,nexons,/*string*/p);
#ifdef DEBUG
	  printf("Unspliced: ");
	  Bitvector_print_file(stdout,retained_introns,/*sep*/',');
	  printf("  Spliced: ");
	  Bitvector_print_file(stdout,spliced_introns,/*sep*/',');
#endif      

	  if (Bitvector_find(retained_introns,all_transcript_introns[transcript_index]) == true) {
	    Uinttableuint_put(transcript_status,transcript_index,UNSPLICED);
	    Uinttable_put(retained_introns_table,transcript_index,(void *) retained_introns);
	    debug(printf(" => UNSPLICED %d %s\n",transcript_index,transcript_name));
	    Bitvector_free(&spliced_introns);
	  } else {
	    debug(printf(" => IGNORE %s in %s\n",transcript_name,transcript_list));
	    Bitvector_free(&retained_introns);
	    Bitvector_free(&spliced_introns);
	  }
	}
      }
      
      FREE(transcript_name);
    }
    
    /* Advance to next transcript */
    while (*p != '\0' && *p != ',') {
      p++;
    }
    if (*p == ',') {
      p++;
    }
  }

  return;
}


static void
make_equiv_classes_block (List_T *equiv_class_lists, List_T lines,
			  int *transcript_nexons, Tableuint_T transcript_dict,
			  List_T *all_transcript_introns,
			  double mean_fragment_length, double robust_sd_fragment_length) {
  bool splicedp = false;
  unsigned int *transcript_indices, *transcript_velocities;
  Bitvector_T *retained_introns;

  Uinttableuint_T transcript_status;
  Uinttable_T retained_introns_table;

  List_T *ptr;
  Equiv_class_T equiv_class;
  int cmp;

  int ntranscripts;

  char *transcript_list;
  unsigned int transcript_index;
  char *line, *auxinfo;

  unsigned int flag;
  int fragment_length;
  List_T l;

  unsigned int *transcript_indices_2;
  int i;


  transcript_status = Uinttableuint_new(/*hint*/10);
  retained_introns_table = Uinttable_new(/*hint*/10);

  /* Parse transcripts */
  for (l = lines; l != NULL; l = List_next(l)) {
    line = (char *) List_head(l);
    auxinfo = Samread_auxinfo(&flag,&fragment_length,line);

    if ((transcript_list = Samread_get_aux_string("XX",auxinfo)) != NULL) {
      debug(printf("XX: %s",line));
      process_transcripts(&splicedp,transcript_status,retained_introns_table,
			  flag,transcript_list,transcript_nexons,transcript_dict,
			  all_transcript_introns,mean_fragment_length,robust_sd_fragment_length);
      debug(printf("\n"));
      FREE(transcript_list);
    }
  }

  if (splicedp == false) {
    for (l = lines; l != NULL; l = List_next(l)) {
      line = (char *) List_head(l);
      auxinfo = Samread_auxinfo(&flag,&fragment_length,line);
    
      if ((transcript_list = Samread_get_aux_string("XY",auxinfo)) != NULL) {
	debug(printf("XY: %s",line));
	process_transcripts(&splicedp,transcript_status,retained_introns_table,
			    flag,transcript_list,transcript_nexons,transcript_dict,
			    all_transcript_introns,mean_fragment_length,robust_sd_fragment_length);
	debug(printf("\n"));
	FREE(transcript_list);
      }
    }
  }

  if ((ntranscripts = Uinttableuint_length(transcript_status)) > 0) {
    transcript_indices = Uinttableuint_keys(transcript_status,/*sortp*/false,/*end*/0);
    transcript_velocities = Uinttableuint_values(transcript_status,/*end*/-1);
    retained_introns = (Bitvector_T *) Uinttable_values(retained_introns_table);
#ifdef CHECK_ASSERTIONS
    transcript_indices_2 = Uinttable_keys(retained_introns_table,/*sortp*/false);
    for (i = 0; i < ntranscripts; i++) {
      assert(transcript_indices[i] == transcript_indices_2[i]);
      assert(retained_introns[i] == NULL ||
	     Bitvector_find(retained_introns[i],all_transcript_introns[transcript_indices[i]]) == true);
    }
    FREE(transcript_indices_2);
#endif
    
    equiv_class = Equiv_class_new(transcript_indices,transcript_velocities,retained_introns,ntranscripts);

    transcript_index = Equiv_class_first_index(equiv_class);
    ptr = &(equiv_class_lists[transcript_index]);

    cmp = -1;
    while (*ptr && (cmp = Equiv_class_cmp((Equiv_class_T) List_head(*ptr),equiv_class)) == -1) {
      ptr = &((*ptr)->rest);
    }

    /* Note: Equiv_class_new sets count to be 1 */
    if (cmp == 0) {
      /* Found existing equiv_class */
      Equiv_class_add_count((Equiv_class_T) List_head(*ptr));
      Equiv_class_free(&equiv_class);

    } else if (*ptr == equiv_class_lists[transcript_index]) {
      /* Insert new equiv_class at the beginning of the list */
      /* Equiv_class_add_count(equiv_class); */
      List_insert(ptr,(void *) equiv_class);
      equiv_class_lists[transcript_index] = *ptr;

    } else {
      /* Insert new equiv_class in the middle of the list */
      /* Equiv_class_add_count(equiv_class); */
      List_insert(ptr,(void *) equiv_class);
    }
  }

  Uinttable_free(&retained_introns_table);
  Uinttableuint_free(&transcript_status);

  return;
}


/* For bulk, equiv classes are stored in an array by their lowest transcript_index */
static List_T *
make_equiv_classes (FILE *fp, int *transcript_nexons, Tableuint_T transcript_dict,
		    List_T *all_transcript_introns, unsigned int n_total_transcripts,
		    double mean_fragment_length, double robust_sd_fragment_length) {
  List_T *equiv_class_lists;

  List_T lines;
  char *line, *acc, *lastacc;
  int linelen;

#ifdef DEBUG
  unsigned int i;
  List_T ptr;
  Equiv_class_T equiv_class;
#endif

  equiv_class_lists = (List_T *) CALLOC(n_total_transcripts,sizeof(List_T));

  lastacc = (char *) MALLOC(1*sizeof(char));
  lastacc[0] = '\0';
  lines = (List_T) NULL;
  while ((line = Getline_wlinefeed(&linelen,fp)) != NULL) {
    acc = get_acc(line);
    if (!strcmp(acc,lastacc)) {
      /* Same acc as before */
      FREE(acc);
      acc = lastacc;
    } else if (lines == NULL) {
      /* First acc seen */
      FREE(lastacc);
      lastacc = acc;
    } else {
      /* Process lines */
      make_equiv_classes_block(equiv_class_lists,lines,transcript_nexons,transcript_dict,
			       all_transcript_introns,mean_fragment_length,robust_sd_fragment_length);
      lines_gc(&lines);
      FREE(lastacc);
      lastacc = acc;
    }
    lines = List_push(lines,(void *) line);
  }

  if (lines == NULL) {
    /* No lines seen */
    FREE(lastacc);
  } else {
    /* Process lines */
    make_equiv_classes_block(equiv_class_lists,lines,transcript_nexons,transcript_dict,
			     all_transcript_introns,mean_fragment_length,robust_sd_fragment_length);
    lines_gc(&lines);
    FREE(lastacc);
  }

#ifdef DEBUG
  for (i = 0; i < n_total_transcripts; i++) {
    for (ptr = equiv_class_lists[i]; ptr != NULL; ptr = List_next(ptr)) {
      equiv_class = (Equiv_class_T) List_head(ptr);
      printf("%d",Equiv_class_count(equiv_class));
      Equiv_class_print(equiv_class);
    }
  }
#endif

  return equiv_class_lists;
}


static void
calc_effective_lengths (double *spliced_lengths, Doublelist_T *unspliced_lengths,
			double mean_fragment_length, int *transcript_lengths,
			int **transcript_exonbounds, Chrpos_T **transcript_exonstarts,
			List_T *all_transcript_introns, unsigned int n_total_transcripts) {
  double effective_length;
  Doublelist_T lengths;
  unsigned int transcript_index;

  List_T p;
  Bitvector_T introns;
  Uintlist_T introns_list, q;
  int introni;
  int *exonbounds, exonlength;
  Chrpos_T *exonstarts, exonend;
  int spliced_length, unspliced_length;
			
  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    spliced_length = transcript_lengths[transcript_index];
    if ((effective_length = (double) spliced_length - mean_fragment_length + 1.0) > 1.0) {
      spliced_lengths[transcript_index] = effective_length;
    } else {
      spliced_lengths[transcript_index] = (double) spliced_length;
    }

    exonstarts = transcript_exonstarts[transcript_index];
    exonbounds = transcript_exonbounds[transcript_index];

    lengths = (Doublelist_T) NULL;
    for (p = all_transcript_introns[transcript_index]; p != NULL; p = List_next(p)) {
      introns = (Bitvector_T) List_head(p);
      introns_list = Bitvector_to_uintlist(introns);

      unspliced_length = spliced_length;
      for (q = introns_list; q != NULL; q = Uintlist_next(q)) {
	introni = Uintlist_head(q);
	if (introni == 0) {
	  exonlength = exonbounds[0];
	} else {
	  exonlength = exonbounds[introni] - exonbounds[introni - 1];
	}

	if (exonstarts[introni+1] > exonstarts[introni]) {
	  /* transcript_genestrand is +1 */
	  exonend = exonstarts[introni] + (Chrpos_T) exonlength - 1;
	  unspliced_length += /*intronlength*/ (int) (exonstarts[introni+1] - exonend) - 1;
	} else {
	  /* transcript_genestrand is -1 */
	  exonend = exonstarts[introni] - (Chrpos_T) exonlength + 1;
	  unspliced_length += /*intronlength*/ (int) (exonend - exonstarts[introni+1]) - 1;
	}
      }

      if ((effective_length = (double) unspliced_length - mean_fragment_length + 1.0) > 1.0) {
	lengths = Doublelist_push(lengths,effective_length);
      } else {
	lengths = Doublelist_push(lengths,(double) unspliced_length);
      }

      Uintlist_free(&introns_list);
    }

    unspliced_lengths[transcript_index] = Doublelist_reverse(lengths);
  }

  return;
}


static void
initialize_weights (List_T *equiv_class_lists,
		    double *spliced_lengths, Doublelist_T *unspliced_lengths,
		    List_T *all_transcript_introns, unsigned int n_total_transcripts) {
  List_T ptr;
  unsigned int i;
  Equiv_class_T equiv_class;
  
  /* Initialize weights */
  for (i = 0; i < n_total_transcripts; i++) {
    for (ptr = equiv_class_lists[i]; ptr != NULL; ptr = List_next(ptr)) {
      equiv_class = (Equiv_class_T) List_head(ptr);
      Equiv_class_calc_weights(equiv_class,spliced_lengths,unspliced_lengths,all_transcript_introns);
    }
  }

  return;
}


#ifdef TWO_STEP_EM
static void
initialize_alpha_spliced (double *alpha_spliced, Doublelist_T *alpha_unspliced,
			  List_T *all_transcript_introns, unsigned int n_total_transcripts) {
  unsigned int transcript_index;
  int n_unspliced, i;
  Doublelist_T values;

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    alpha_spliced[transcript_index] = 1.0/(double) n_total_transcripts;

    if ((n_unspliced = List_length(all_transcript_introns[transcript_index])) == 0) {
      alpha_unspliced[transcript_index] = (Doublelist_T) NULL;

    } else {
      values = (Doublelist_T) NULL;
      for (i = 0; i < n_unspliced; i++) {
	values = Doublelist_push(values,0.0);
      }
      alpha_unspliced[transcript_index] = values;
    }
  }

  return;
}


static void
initialize_alpha_unspliced (double *alpha_spliced, Doublelist_T *alpha_unspliced,
			    List_T *all_transcript_introns, unsigned int n_total_transcripts) {
  unsigned int transcript_index;
  int n_unspliced;
  double value;
  Doublelist_T p;

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    if ((n_unspliced = List_length(all_transcript_introns[transcript_index])) == 0) {
      alpha_unspliced[transcript_index] = (Doublelist_T) NULL;

    } else if (alpha_spliced[transcript_index] > 0.0) {
      n_unspliced = List_length(all_transcript_introns[transcript_index]);

      value = alpha_spliced[transcript_index] / (double) (1 + n_unspliced);
      alpha_spliced[transcript_index] = value;
      for (p = alpha_unspliced[transcript_index]; p != NULL; p = Doublelist_next(p)) {
	Doublelist_head_set(p,value);
      }

    } else {
      /* Leave alpha_unspliced as a Doublelist_T of zeroes */
    }
  }

  return;
}

#else

static void
initialize_alpha (double *alpha_spliced, Doublelist_T *alpha_unspliced,
		  List_T *all_transcript_introns, unsigned int n_total_transcripts) {
  unsigned int transcript_index;
  int n_unspliced, i;
  double value;
  Doublelist_T values;

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    if ((n_unspliced = List_length(all_transcript_introns[transcript_index])) == 0) {
      alpha_spliced[transcript_index] = 1.0/(double) n_total_transcripts;
      alpha_unspliced[transcript_index] = (Doublelist_T) NULL;

    } else {
      alpha_spliced[transcript_index] = value = 1.0/(double) n_total_transcripts/(double) (1 + n_unspliced);

      values = (Doublelist_T) NULL;
      for (i = 0; i < n_unspliced; i++) {
	values = Doublelist_push(values,value);
      }
      alpha_unspliced[transcript_index] = values;
    }
  }

  return;
}
#endif


static Doublelist_T *
make_alpha_unspliced (List_T *all_transcript_introns, unsigned int n_total_transcripts) {
  Doublelist_T *alpha_unspliced, values;
  unsigned int transcript_index;
  int n_unspliced, i;

  alpha_unspliced = (Doublelist_T *) CALLOC(n_total_transcripts,sizeof(Doublelist_T));

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    if ((n_unspliced = List_length(all_transcript_introns[transcript_index])) == 0) {
      alpha_unspliced[transcript_index] = (Doublelist_T) NULL;

    } else {
      values = (Doublelist_T) NULL;
      for (i = 0; i < n_unspliced; i++) {
	values = Doublelist_push(values,0.0);
      }
      alpha_unspliced[transcript_index] = values;
    }
  }

  return alpha_unspliced;
}



static void
EM_iteration (double **next_alpha_spliced, Doublelist_T **next_alpha_unspliced,
	      List_T *equiv_class_lists, double *alpha_spliced, Doublelist_T *alpha_unspliced,
	      List_T *all_transcript_introns, unsigned int n_total_transcripts) {
  List_T ptr;
  Equiv_class_T equiv_class;
  unsigned int transcript_index;

  *next_alpha_spliced = (double *) CALLOC(n_total_transcripts,sizeof(double));
  *next_alpha_unspliced = make_alpha_unspliced(all_transcript_introns,n_total_transcripts);

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    for (ptr = equiv_class_lists[transcript_index]; ptr != NULL; ptr = List_next(ptr)) {
      equiv_class = (Equiv_class_T) List_head(ptr);
      Equiv_class_EM_update(*next_alpha_spliced,*next_alpha_unspliced,equiv_class,
			    alpha_spliced,alpha_unspliced,all_transcript_introns);
    }
  }

  return;
}
  

/* Taken from Kallisto */
#define MIN_ROUNDS 50
#define ALPHA_LIMIT 1e-7
#define ALPHA_CHANGE_LIMIT 0.01
#define ALPHA_CHANGE 0.01

static void
perform_EM (double **alpha_spliced, Doublelist_T **alpha_unspliced,
	    List_T *equiv_class_lists, List_T *all_transcript_introns,
	    unsigned int n_total_transcripts) {
  unsigned int transcript_index;

  double *next_alpha_spliced;
  Doublelist_T *next_alpha_unspliced, p, q;
  int iter;

  int nchanged;
  

  /* Perform iterations */
  iter = 0;
  nchanged = (int) n_total_transcripts;
  while (iter < 10000 && (iter <= MIN_ROUNDS || nchanged > 0)) {
    EM_iteration(&next_alpha_spliced,&next_alpha_unspliced,equiv_class_lists,
		 *alpha_spliced,*alpha_unspliced,all_transcript_introns,n_total_transcripts);

    nchanged = 0;
    for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
      if (next_alpha_spliced[transcript_index] > ALPHA_CHANGE_LIMIT &&
	  fabs(next_alpha_spliced[transcript_index] - (*alpha_spliced)[transcript_index]) / next_alpha_spliced[transcript_index] > ALPHA_CHANGE) {
	nchanged += 1;
      }

      for (p = next_alpha_unspliced[transcript_index], q = (*alpha_unspliced)[transcript_index]; p != NULL;
	   p = Doublelist_next(p), q = Doublelist_next(q)) {
	if (Doublelist_head(p) > ALPHA_CHANGE_LIMIT &&
	    fabs(Doublelist_head(p) - Doublelist_head(q)) / Doublelist_head(p) > ALPHA_CHANGE) {
	  nchanged += 1;
	}
      }
      Doublelist_free(&((*alpha_unspliced)[transcript_index]));
    }
    
    FREE(*alpha_spliced);
    FREE(*alpha_unspliced);

    *alpha_spliced = next_alpha_spliced;
    *alpha_unspliced = next_alpha_unspliced;
    iter++;
  }
  fprintf(stderr,"Performed %d iterations\n",iter);
    
  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    if ((*alpha_spliced)[transcript_index] < ALPHA_LIMIT/10.0) {
      (*alpha_spliced)[transcript_index] = 0.0;
    }
    for (p = (*alpha_unspliced)[transcript_index]; p != NULL; p = Doublelist_next(p)) {
      if (Doublelist_head(p) < ALPHA_LIMIT/10.0) {
	Doublelist_head_set(p,0.0);
      }
    }
  }

  return;
}


static void
print_alpha (double *alpha_spliced, Doublelist_T *alpha_unspliced,
	     int *transcript_lengths, double *spliced_lengths,
	     List_T *all_transcript_introns, char **all_transcripts, char **all_genes,
	     unsigned int *transcripti_to_genei, unsigned int n_total_transcripts) {
  unsigned int transcript_index;
  unsigned int gene_index;
  List_T p;
  Doublelist_T q;
  double alpha_unspliced_total;
  
  
  debug(printf("index\t"));
  printf("transcript\tgene\t");
  printf("transcript.len\tspliced.efflen\t");
  printf("spliced.count\tunspliced.count\t");
  printf("retained.introns\n");
  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    alpha_unspliced_total = Doublelist_sum(alpha_unspliced[transcript_index]);

    if (nonzero_only_p == true && alpha_spliced[transcript_index] == 0 && alpha_unspliced_total == 0) {
      /* Skip */
    } else {
      gene_index = transcripti_to_genei[transcript_index]; /* 0-based */

      debug(printf("%u\t",i));
      printf("%s\t%s\t",all_transcripts[transcript_index],all_genes[gene_index]);
      printf("%d\t%.2f\t",transcript_lengths[transcript_index],spliced_lengths[transcript_index]);
      printf("%f\t%f\t",alpha_spliced[transcript_index],alpha_unspliced_total);

      if (all_transcript_introns[transcript_index] == (List_T) NULL) {
	printf("NA");
      } else {
	p = all_transcript_introns[transcript_index];
	q = alpha_unspliced[transcript_index];
	Bitvector_print_file(stdout,(Bitvector_T) List_head(p),/*sep*/',');
	printf(":%f",Doublelist_head(q));
	for (p = List_next(p), q = Doublelist_next(q); p != NULL; p = List_next(p), q = Doublelist_next(q)) {
	  printf(";");
	  Bitvector_print_file(stdout,(Bitvector_T) List_head(p),/*sep*/',');
	  printf(":%f",Doublelist_head(q));
	}
      }

      printf("\n");
    }
  }
  
  return;
}


static int
parse_command_line (int argc, char *argv[], int optind) {
  int opt, c;
  extern char *optarg;
  int long_option_index = 0;
  const char *long_name;
  char **argstart;
  /* char *string; */

  fprintf(stderr,"bulk-counts version %s called with args:",PACKAGE_VERSION);
  argstart = &(argv[-optind]);
  for (c = 1; c < argc + optind; c++) {
      fprintf(stderr," %s",argstart[c]);
  }
  fprintf(stderr,"\n");

  while ((opt = getopt_long(argc,argv,"D:d:C:c:o:",
			    long_options, &long_option_index)) != -1) {
    switch (opt) {
    case 0:
      long_name = long_options[long_option_index].name;
      if (!strcmp(long_name,"version")) {
	print_program_version();
	return 1;

      } else if (!strcmp(long_name,"nonzero-only")) {
	nonzero_only_p = true;

      } else if (!strcmp(long_name,"help")) {
	print_program_usage();
	return 1;

      } else {
	/* Shouldn't reach here */
	fprintf(stderr,"Don't recognize option %s.  For usage, run 'gsnap --help'",long_name);
	return 9;
      }
      break;

    case 'D': user_genomedir = optarg; break;
    case 'd': genome_dbroot = optarg; break;
    case 'C': user_transcriptomedir = optarg; break;
    case 'c': transcriptome_dbroot = optarg; break;

    case 'o': output_prefix = optarg; break;

    case '?': fprintf(stderr,"For usage, run 'gsnap --help'\n"); return 9;
    default: return 9;
    }
  }

  if (genome_dbroot == NULL) {
    fprintf(stderr,"Need to specify the -d flag.  For usage, run 'velocity-counts --help'\n");
    /* print_program_usage(); */
    return 9;
  }

  if (transcriptome_dbroot == NULL) {
    fprintf(stderr,"Need to specify the -c flag.  For usage, run 'velocity-counts --help'\n");
    /* print_program_usage(); */
    return 9;
  }

  return 0;
}


int
main (int argc, char *argv[]) {
  int cmdline_status;

  char *genomesubdir = NULL, *genome_fileroot = NULL, *dbversion;
  char *transcriptomesubdir = NULL, *transcriptome_fileroot = NULL;
  IIT_T transcript_map_iit;
  Transcriptome_T transcriptome;

  Equiv_class_T equiv_class;
  int *transcript_lengths;
  double *spliced_lengths;
  Doublelist_T *unspliced_lengths;
  int *transcript_nexons, **transcript_exonbounds;
  Chrpos_T **transcript_exonstarts;
  List_T *all_transcript_introns, ptr;
  Bitvector_T introns;
  double mean_fragment_length, sd_fragment_length, median_fragment_length, robust_sd_fragment_length;

  Tableuint_T transcript_dict;	/* transcript_name -> index (0-based) */

  /* Expecting GSNAP output from transcriptome-guided genomic alignment */
  char *alignment_file, *iitfile;
  FILE *fp;

  char **all_transcripts, **all_genes;
  unsigned int *transcripti_to_genei;
  Uintlist_T *genei_to_transcripti, transcript_list;
  unsigned int n_total_transcripts, n_total_genes, transcript_index, gene_index;

  double *alpha_spliced;
  Doublelist_T *alpha_unspliced;
  List_T *equiv_class_lists;


  cmdline_status = parse_command_line(argc,argv,optind);
  argc -= optind;
  argv += optind;

  if (cmdline_status == 0) {
    /* okay to continue */
  } else if (cmdline_status == 1) {
    exit(0);
  } else {
    exit(cmdline_status);
  }

  /* Read transcripts and symbols */
  genomesubdir = Datadir_find_genomesubdir(&genome_fileroot,&dbversion,user_genomedir,genome_dbroot);
  FREE(dbversion);

  if (user_transcriptomedir == NULL && user_genomedir != NULL) {
    user_transcriptomedir = user_genomedir;
  }
  transcriptomesubdir = Datadir_find_genomesubdir(&transcriptome_fileroot,&dbversion,user_transcriptomedir,
						  transcriptome_dbroot);
  FREE(dbversion);

  /* transcript_map_iit relates transcripts to genomic coordinates and is an IIT_T */
  /* Use transcript_map_iit because it has gene symbols */
  iitfile = (char *) CALLOC(strlen(genomesubdir)+strlen("/")+
			    strlen(genome_fileroot)+strlen(".transcripts/")+
			    strlen(transcriptome_fileroot)+strlen(".align.iit")+1,sizeof(char));
  sprintf(iitfile,"%s/%s.transcripts/%s.align.iit",genomesubdir,genome_fileroot,transcriptome_fileroot);

  if ((transcript_map_iit = IIT_read(iitfile,/*name*/transcriptome_dbroot,/*readonlyp*/true,/*divread*/READ_ALL,
				     /*divstring*/NULL,/*add_iit_p*/false)) == NULL) {
    fprintf(stderr,"Transcript map file %s not found\n",iitfile);
    exit(9);
  } else {
    transcript_dict = get_transcripts(&n_total_transcripts,transcript_map_iit);
    all_transcripts = (char **) Tableuint_keys_by_timeindex(transcript_dict,NULL);
    all_genes = get_genes(&n_total_genes,&transcripti_to_genei,&genei_to_transcripti,
			  transcript_map_iit);
    IIT_free(&transcript_map_iit);
    FREE(iitfile);
  }

  /* Read transcriptome, which has exon info */
  transcriptome = Transcriptome_new(genomesubdir,genome_fileroot,transcriptome_fileroot,/*sharedp*/false);
  transcript_lengths = get_exon_info(&transcript_nexons,&transcript_exonbounds,&transcript_exonstarts,
				     transcriptome,n_total_transcripts);
  /* Cannot free transcriptome here, since exonbounds and exonstarts point to it */

  FREE(genomesubdir);
  FREE(genome_fileroot);
  FREE(transcriptomesubdir);
  FREE(transcriptome_fileroot);


  /* Pass 1: Find retained introns */
  alignment_file = argv[0];
  if ((fp = fopen(alignment_file,"r")) == NULL) {
    fprintf(stderr,"Could not open %s\n",alignment_file);
    exit(9);
  } else {
    all_transcript_introns = pass1(&median_fragment_length,&robust_sd_fragment_length,
				   &mean_fragment_length,&sd_fragment_length,
				   fp,transcript_nexons,transcript_dict,all_transcripts,
				   n_total_transcripts);
    fclose(fp);

    fprintf(stderr,"Mean fragment length is %.2f\n",mean_fragment_length);
    fprintf(stderr,"SD of fragment length is %.2f\n",sd_fragment_length);
    fprintf(stderr,"Median fragment length is %d\n",(int) median_fragment_length);
    fprintf(stderr,"Robust sd is %.2f\n",robust_sd_fragment_length);
  }


  /* Pass 2: Make equiv classes */
  if ((fp = fopen(alignment_file,"r")) == NULL) {
    fprintf(stderr,"Could not open %s\n",alignment_file);
    exit(9);
  } else {
    equiv_class_lists =
      make_equiv_classes(fp,transcript_nexons,transcript_dict,all_transcript_introns,n_total_transcripts,
			 mean_fragment_length,robust_sd_fragment_length);
    fclose(fp);
  }


  /* Calculate effective transcript lengths */
  spliced_lengths = (double *) MALLOC(n_total_transcripts*sizeof(double));
  unspliced_lengths = (Doublelist_T *) MALLOC(n_total_transcripts*sizeof(Doublelist_T));
  calc_effective_lengths(spliced_lengths,unspliced_lengths,
			 mean_fragment_length,transcript_lengths,
			 transcript_exonbounds,transcript_exonstarts,
			 all_transcript_introns,n_total_transcripts);
  initialize_weights(equiv_class_lists,spliced_lengths,unspliced_lengths,
		     all_transcript_introns,n_total_transcripts);

  FREE(transcript_nexons);
  FREE(transcript_exonbounds);
  FREE(transcript_exonstarts);
  Transcriptome_free(&transcriptome);

  alpha_spliced = (double *) MALLOC(n_total_transcripts*sizeof(double));
  alpha_unspliced = (Doublelist_T *) MALLOC(n_total_transcripts*sizeof(Doublelist_T));

#ifdef TWO_STEP_EM
  initialize_alpha_spliced(alpha_spliced,alpha_unspliced,all_transcript_introns,n_total_transcripts);
  perform_EM(&alpha_spliced,&alpha_unspliced,equiv_class_lists,
	     all_transcript_introns,n_total_transcripts);

  initialize_alpha_unspliced(alpha_spliced,alpha_unspliced,all_transcript_introns,n_total_transcripts);
  perform_EM(&alpha_spliced,&alpha_unspliced,equiv_class_lists,
	     all_transcript_introns,n_total_transcripts);
#else
  initialize_alpha(alpha_spliced,alpha_unspliced,all_transcript_introns,n_total_transcripts);
  perform_EM(&alpha_spliced,&alpha_unspliced,equiv_class_lists,
	     all_transcript_introns,n_total_transcripts);
#endif

  print_alpha(alpha_spliced,alpha_unspliced,transcript_lengths,spliced_lengths,/*unspliced_lengths,*/
	      all_transcript_introns,all_transcripts,all_genes,transcripti_to_genei,n_total_transcripts);

  /* Free memory */
  FREE(alpha_spliced);
  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    Doublelist_free(&(alpha_unspliced[transcript_index]));
  }
  FREE(alpha_unspliced);

  FREE(spliced_lengths);
  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    Doublelist_free(&(unspliced_lengths[transcript_index]));
  }
  FREE(unspliced_lengths);
  FREE(transcript_lengths);

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    for (ptr = all_transcript_introns[transcript_index]; ptr != NULL; ptr = List_next(ptr)) {
      introns = (Bitvector_T) List_head(ptr);
      Bitvector_free(&introns);
    }
    List_free(&(all_transcript_introns[transcript_index]));
  }
  FREE(all_transcript_introns);

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    for (ptr = equiv_class_lists[transcript_index]; ptr != NULL; ptr = List_next(ptr)) {
      equiv_class = (Equiv_class_T) List_head(ptr);
      Equiv_class_free(&equiv_class);
    }
    List_free(&(equiv_class_lists[transcript_index]));
  }
  FREE(equiv_class_lists);

  FREE(transcripti_to_genei);

  for (gene_index = 0; gene_index < n_total_genes; gene_index++) {
    transcript_list = genei_to_transcripti[gene_index];
    Uintlist_free(&transcript_list);
    FREE(all_genes[gene_index]);
  }
  FREE(all_genes);
  FREE(genei_to_transcripti);

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    FREE(all_transcripts[transcript_index]);
  }
  FREE(all_transcripts);
  Tableuint_free(&transcript_dict);

  return 0;
}


 static void
print_program_usage () {
  fprintf(stdout,"\
Usage: bulk-counts [OPTIONS...] <GSNAP alignment file>\n\
\n\
");

  /* Input options */
  fprintf(stdout,"Options (must include -d and -c)\n");
  fprintf(stdout,"\
  -D, --dir=directory            Genome directory.  Default (as specified by --with-gmapdb to the configure program) is\n\
                                   %s\n\
",GMAPDB);
  fprintf(stdout,"\
  -d, --db=STRING                Genome database\n\
\n\
");

  fprintf(stdout,"\
  -C, --transcriptdir=directory  Transcriptome directory.  Default is the value for --dir above\n\
  -c, --transcriptdb=STRING      Transcriptome database\n\
\n\
");

  fprintf(stdout,"\
  --input-command                Process alignment file using this command (e.g., \"samtools view\" or \"gunzip -c\")\n\
  -o, --output=STRING            Prefix for output files.  Can be a directory if it ends with '/'\n\
  --nonzero-only                 Print only transcripts and genes with nonzero results\n\
\n\
");

  return;
 }
