static char rcsid[] = "$Id$";
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "equiv-class-bulk.h"
#include <stdio.h>		/* For sprintf */
#include <stdlib.h>
#include <string.h>		/* For strlen */

#if defined(HAVE_POPCNT)
#include <immintrin.h>
#elif defined(HAVE_MM_POPCNT)
#include <nmmintrin.h>
#endif

#include "assert.h"
#include "mem.h"
#include "types.h"		/* For Genomecomp_T */
#include "sedgesort.h"
#include "uintlist.h"
#include "doublelist.h"


static char velocity_char[4] = "NBSU";


#define T Equiv_class_T
struct T {

  int ntranscripts;
  unsigned int *transcripts;	/* indices */
  unsigned int *velocities;	/* SPLICED, UNSPLICED, BOTH */
  Bitvector_T *retained_introns; /* one per transcript */

  double *spliced_weights;
  Doublelist_T *unspliced_weights;

  int count;			/* For bulk, count is nreads.  For single cell, count is always 1 */
};

void
Equiv_class_free (T *old) {
  int i;

  FREE((*old)->transcripts);
  FREE((*old)->velocities);

  for (i = 0; i < (*old)->ntranscripts; i++) {
    Bitvector_free(&((*old)->retained_introns[i]));
  }
  FREE((*old)->retained_introns);

  FREE((*old)->spliced_weights);

  if ((*old)->unspliced_weights != NULL) {
    for (i = 0; i < (*old)->ntranscripts; i++) {
      Doublelist_free(&((*old)->unspliced_weights[i]));
    }
    FREE((*old)->unspliced_weights);
  }

  FREE(*old);

  return;
}


/* transcript_indices and transcript_velocities have (n+1) slots, from Tableint procedures */
T
Equiv_class_new (unsigned int *transcript_indices, unsigned int *transcript_velocities,
		 Bitvector_T *retained_introns, int ntranscripts) {
  T new = (T) MALLOC(sizeof(*new));
  int *order;
  int i;

  new->ntranscripts = ntranscripts;
  order = Sedgesort_order(transcript_indices,ntranscripts);

  new->transcripts = (unsigned int *) MALLOC(new->ntranscripts*sizeof(unsigned int));
  new->velocities = (unsigned int *) MALLOC(new->ntranscripts*sizeof(unsigned int));
  new->retained_introns = (Bitvector_T *) MALLOC(new->ntranscripts*sizeof(Bitvector_T));

  for (i = 0; i < new->ntranscripts; i++) {
    new->transcripts[i] = transcript_indices[order[i]];
    new->velocities[i] = transcript_velocities[order[i]];
    new->retained_introns[i] = retained_introns[order[i]];
  }

  /* Allocate after all revisions, if any */
  new->spliced_weights = (double *) NULL;
  new->unspliced_weights = (Doublelist_T *) NULL;

  new->count = 1;

  FREE(order);
  FREE(retained_introns);
  FREE(transcript_velocities);
  FREE(transcript_indices);

  return new;
}

/* Specific to bulk */
unsigned int
Equiv_class_first_index (T this) {
  return this->transcripts[0];
}

int
Equiv_class_count (T this) {
  return this->count;
}

void
Equiv_class_add_count (T this) {
  this->count += 1;
  return;
}

void
Equiv_class_print (T this) {
  int i;

  for (i = 0; i < this->ntranscripts; i++) {
    printf(" %u",this->transcripts[i]);
    if (this->velocities[i] == SPLICED) {
      printf("S");
    } else if (this->velocities[i] == UNSPLICED) {
      printf("U");
    } else if (this->velocities[i] == BOTH) {
      printf("B");
    }
    if (Bitvector_null_p(this->retained_introns[i]) == false) {
      printf("[");
      Bitvector_print_file(stdout,this->retained_introns[i],/*sep*/',');
      printf("]");
    }
  }
  printf("\n");
}

/* Equiv_class_cmp is called on the structures and not pointers to structures */
int
Equiv_class_cmp (const void *a, const void *b) {
  T x = (T) a;
  T y = (T) b;
  int i = 0;
  int cmp;
  
  if (x->ntranscripts < y->ntranscripts) {
    while (i < x->ntranscripts) {
      if (x->transcripts[i] < y->transcripts[i]) {
	return -1;
      } else if (y->transcripts[i] < x->transcripts[i]) {
	return +1;
      } else if (x->velocities[i] < y->velocities[i]) {
	return -1;
      } else if (y->velocities[i] < x->velocities[i]) {
	return +1;
      } else if ((cmp = Bitvector_cmp(x->retained_introns[i],y->retained_introns[i])) < 0) {
	return -1;
      } else if (cmp > 0) {
	return +1;
      } else {
	i++;
      }
    }
    return -1;			/* x has a shorter list */

  } else if (y->ntranscripts < x->ntranscripts) {
    while (i < y->ntranscripts) {
      if (x->transcripts[i] < y->transcripts[i]) {
	return -1;
      } else if (y->transcripts[i] < x->transcripts[i]) {
	return +1;
      } else if (x->velocities[i] < y->velocities[i]) {
	return -1;
      } else if (y->velocities[i] < x->velocities[i]) {
	return +1;
      } else if ((cmp = Bitvector_cmp(x->retained_introns[i],y->retained_introns[i])) < 0) {
	return -1;
      } else if (cmp > 0) {
	return +1;
      } else {
	i++;
      }
    }
    return +1;			/* y has a shorter list */
    
  } else {
    while (i < x->ntranscripts) {
      if (x->transcripts[i] < y->transcripts[i]) {
	return -1;
      } else if (y->transcripts[i] < x->transcripts[i]) {
	return +1;
      } else if (x->velocities[i] < y->velocities[i]) {
	return -1;
      } else if (y->velocities[i] < x->velocities[i]) {
	return +1;
      } else if ((cmp = Bitvector_cmp(x->retained_introns[i],y->retained_introns[i])) < 0) {
	return -1;
      } else if (cmp > 0) {
	return +1;
      } else {
	i++;
      }
    }
    return 0;			/* both lists are equal */
  }
}


void
Equiv_class_calc_weights (T this, double *spliced_lengths, Doublelist_T *unspliced_lengths,
			  List_T *all_transcript_introns) {
  int i;
  unsigned int transcript_index;
  unsigned int velocity_type;
  Doublelist_T p;
  double unspliced_length;

  if (this->ntranscripts > 0) {
    this->spliced_weights = (double *) CALLOC(this->ntranscripts,sizeof(double));
    this->unspliced_weights = (Doublelist_T *) CALLOC(this->ntranscripts,sizeof(Doublelist_T));

    for (i = 0; i < this->ntranscripts; i++) {
      transcript_index = this->transcripts[i];
      
      if ((velocity_type = this->velocities[i]) == SPLICED) {
	this->spliced_weights[i] = (double) this->count / spliced_lengths[transcript_index];
	this->unspliced_weights[i] = (Doublelist_T) NULL;

      } else if (velocity_type == UNSPLICED) {
	this->spliced_weights[i] = 0.0;
	unspliced_length = Bitvector_double_find(this->retained_introns[i],all_transcript_introns[transcript_index],
						 unspliced_lengths[transcript_index]);
	this->unspliced_weights[i] = Doublelist_push(NULL,this->count / unspliced_length);

      } else if (velocity_type == BOTH) {
	this->spliced_weights[i] = (double) this->count / spliced_lengths[transcript_index];
	this->unspliced_weights[i] = (Doublelist_T) NULL;
	for (p = unspliced_lengths[transcript_index]; p != NULL; p = Doublelist_next(p)) {
	  unspliced_length = Doublelist_head(p);
	  this->unspliced_weights[i] = Doublelist_push(this->unspliced_weights[i],this->count / unspliced_length);
	}
	this->unspliced_weights[i] = Doublelist_reverse(this->unspliced_weights[i]);
      }
    }
  }

  return;
}


#define TOLERANCE 2.22045e-15

/* For BOTH, treat all spliced and unspliced transcripts as candidates */
void
Equiv_class_EM_update (double *next_alpha_spliced, Doublelist_T *next_alpha_unspliced, T this,
		       double *alpha_spliced, Doublelist_T *alpha_unspliced, List_T *all_transcript_introns) {
  double denom = 0.0, count_norm, alpha;
  unsigned int transcript_index;
  unsigned int velocity_type;
  Doublelist_T p, nextp, q, ptr;
  int i;

  assert(this->count > 0);

  for (i = 0; i < this->ntranscripts; i++) {
    transcript_index = this->transcripts[i];
    
    if ((velocity_type = this->velocities[i]) == SPLICED) {
      denom += alpha_spliced[transcript_index] * this->spliced_weights[i];
      
    } else if (velocity_type == UNSPLICED) {
      alpha = Bitvector_double_find(this->retained_introns[i],all_transcript_introns[transcript_index],
				    alpha_unspliced[transcript_index]);
      denom += alpha * Doublelist_head(this->unspliced_weights[i]);
      
    } else if (velocity_type == BOTH) {
      denom += alpha_spliced[transcript_index] * this->spliced_weights[i];
      for (p = alpha_unspliced[transcript_index], q = this->unspliced_weights[i];
	   p != NULL; p = Doublelist_next(p), q = Doublelist_next(q)) {
	denom += Doublelist_head(p) * Doublelist_head(q);
      }
    }
  }
  
  if (denom >= TOLERANCE) {
    count_norm = (double) this->count / denom;
    for (i = 0; i < this->ntranscripts; i++) {
      transcript_index = this->transcripts[i];
      if ((velocity_type = this->velocities[i]) == SPLICED) {
	next_alpha_spliced[transcript_index] +=
	  alpha_spliced[transcript_index] * this->spliced_weights[i] * count_norm;
	
      } else if (velocity_type == UNSPLICED) {
	alpha = Bitvector_double_find(this->retained_introns[i],all_transcript_introns[transcript_index],
				      alpha_unspliced[transcript_index]);
	ptr = Bitvector_double_find_ptr(this->retained_introns[i],all_transcript_introns[transcript_index],
					next_alpha_unspliced[transcript_index]);
	Doublelist_head_add(ptr,alpha * Doublelist_head(this->unspliced_weights[i]) * count_norm);
	
      } else if (velocity_type == BOTH) {
	next_alpha_spliced[transcript_index] +=
	  alpha_spliced[transcript_index] * this->spliced_weights[i] * count_norm;
	
	for (p = alpha_unspliced[transcript_index], nextp = next_alpha_unspliced[transcript_index],
	       q = this->unspliced_weights[i]; p != NULL;
	     p = Doublelist_next(p), nextp = Doublelist_next(nextp), q = Doublelist_next(q)) {
	  Doublelist_head_add(nextp,Doublelist_head(p) * Doublelist_head(q) * count_norm);
	}
      }
    }
  }

  return;
}

