static char rcsid[] = "$Id$";
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "equiv-class-sc.h"
#include <stdio.h>		/* For sprintf */
#include <stdlib.h>
#include <string.h>		/* For strlen */

#if defined(HAVE_POPCNT)
#include <immintrin.h>
#elif defined(HAVE_MM_POPCNT)
#include <nmmintrin.h>
#endif

#include "assert.h"
#include "mem.h"
#include "types.h"		/* For Genomecomp_T */
#include "sedgesort.h"
#include "uintlist.h"
#include "popcount.h"


static char velocity_char[4] = "NBSU";


#define T Equiv_class_T
struct T {
  int umi_id; /* For distinguishing transcripts with the same umi
		 barcode.  Also, a value of -1 indicates the equiv
		 class has been merged into another one */
  int ntranscripts;
  unsigned int *transcripts;	/* indices */
  unsigned int *velocities;	/* SPLICED, UNSPLICED, BOTH */
  Bitvector_T *retained_introns; /* one per transcript */
  Bitvector_T *spliced_introns;	/* one per transcript */

  double *spliced_weights;
  Doublelist_T *unspliced_weights;

  int ngenes;
  unsigned int *genes;
  Genomecomp_T high, low, flags; /* For the UMI barcode */
  
  int count;			/* For bulk, count is nreads.  For single cell, count is always 1 */
  int nreads;			/* Specific to single cell */
};

void
Equiv_class_free (T *old) {
  int i;

  FREE((*old)->genes);
  FREE((*old)->transcripts);
  FREE((*old)->velocities);

  for (i = 0; i < (*old)->ntranscripts; i++) {
    Bitvector_free(&((*old)->retained_introns[i]));
    Bitvector_free(&((*old)->spliced_introns[i]));
  }
  FREE((*old)->retained_introns);
  FREE((*old)->spliced_introns);

  FREE((*old)->spliced_weights);
  FREE((*old)->unspliced_weights);
  FREE(*old);
  return;
}


/*                    87654321 */
#define RIGHT_SET   0x00000001
#define RIGHT_CLEAR 0x00000000


/* Genome128 format */
/*          High bit   Low bit   Flag bit
   A = 000     0          0         0
   C = 001     0          1         0
   G = 010     1          0         0
   T = 011     1          1         1
   N = 100     0          0         1
   X = 111     1          1         1
*/

/* Specific to single cell */
static void
compress_barcode (Genomecomp_T *high, Genomecomp_T *low, Genomecomp_T *flags,
		  char *barcode) {
  int pos, length;
  char c;

  length = strlen(barcode);
  assert(length <= 32);
  
  *high = *low = *flags = 0U;
  for (pos = 0; pos < length; pos++) {
    c = barcode[pos];
    *high <<= 1;
    *low <<= 1;
    *flags <<= 1;

    /* Assume that barcode is upper case */
    switch /*(uppercaseCode[c])*/ (c) {
    case 'A': /* *high |= RIGHT_CLEAR; */ /* *low |= RIGHT_CLEAR; */ /* *flags |= RIGHT_CLEAR; */ break;
    case 'C': /* *high |= RIGHT_CLEAR; */    *low |= RIGHT_SET;      /* *flags |= RIGHT_CLEAR; */ break;
    case 'G':    *high |= RIGHT_SET;      /* *low |= RIGHT_CLEAR; */ /* *flags |= RIGHT_CLEAR; */ break;
    case 'T':    *high |= RIGHT_SET;         *low |= RIGHT_SET;      /* *flags |= RIGHT_CLEAR; */ break;
    default:  /* *high |= RIGHT_CLEAR; */ /* *low |= RIGHT_CLEAR; */    *flags |= RIGHT_SET;
    }
  }

  return;
}


static void
print_barcode (FILE *fp, Genomecomp_T high, Genomecomp_T low, Genomecomp_T flags,
	       int barcode_len) {
  char Buffer[33];
  int i, k;

  /* printf("%08X %08X %08X => ",high,low,flags); */

  k = barcode_len;
  for (i = 0; i < barcode_len; i++) {
    switch (((high & 0x01) << 1) | (low & 0x01)) {
    case 0U: Buffer[--k] = 'A'; break;
    case 1U: Buffer[--k] = 'C'; break;
    case 2U: Buffer[--k] = 'G'; break;
    case 3U: Buffer[--k] = 'T'; break;
    default: abort();
    }
    high >>= 1;
    low >>= 1;
  }

  if (flags != 0U) {
    k = barcode_len;
    for (i = 0; i < barcode_len; i++) {
      if (flags & 0x01) {
	Buffer[--k] = 'N';
      }
      flags >>= 1;
    }
  }

  Buffer[barcode_len] = '\0';

  fprintf(fp,"%s",Buffer);
  return;
}


void
Equiv_class_print_barcode (FILE *fp, T this) {
  print_barcode(fp,this->high,this->low,this->flags,/*barcode_len*/12);
  return;
}


/* transcript_indices and transcript_velocities have (n+1) slots, from Tableint procedures */
T
Equiv_class_new (unsigned int *transcript_indices, unsigned int *transcript_velocities,
		 Bitvector_T *retained_introns, Bitvector_T *spliced_introns, int ntranscripts,
		 char *umi_barcode, int umi_id) {
  T new = (T) MALLOC(sizeof(*new));
  int *order;
  int i;

  new->umi_id = umi_id;
  new->ntranscripts = ntranscripts;
  order = Sedgesort_order(transcript_indices,ntranscripts);

  new->transcripts = (unsigned int *) MALLOC(new->ntranscripts*sizeof(unsigned int));
  new->velocities = (unsigned int *) MALLOC(new->ntranscripts*sizeof(unsigned int));
  new->retained_introns = (Bitvector_T *) MALLOC(new->ntranscripts*sizeof(Bitvector_T));
  new->spliced_introns = (Bitvector_T *) MALLOC(new->ntranscripts*sizeof(Bitvector_T));

  new->ngenes = 0;
  new->genes = (unsigned int *) NULL;

  for (i = 0; i < new->ntranscripts; i++) {
    new->transcripts[i] = transcript_indices[order[i]];
    new->velocities[i] = transcript_velocities[order[i]];
    new->retained_introns[i] = retained_introns[order[i]];
    new->spliced_introns[i] = spliced_introns[order[i]];
  }

  /* Allocate after all revisions, if any */
  new->spliced_weights = (double *) NULL;
  new->unspliced_weights = (Doublelist_T *) NULL;

  if (umi_barcode == NULL) {
    new->high = new->low = new->flags = 0U;
  } else {
    compress_barcode(&new->high,&new->low,&new->flags,umi_barcode);
  }

  new->nreads = 1;

  FREE(order);
  FREE(spliced_introns);
  FREE(retained_introns);
  FREE(transcript_velocities);
  FREE(transcript_indices);

  return new;
}

/* Specific to single cell */
int
Equiv_class_umi_id (T this) {
  return this->umi_id;
}

void
Equiv_class_set_umi_id (T this, int umi_id) {
  this->umi_id = umi_id;
}


/* Specific to single cell */
/* Takes the intersection of old and new transcripts */
/* Revises intron_status for each transcript, by adding introns, and revising BOTH to SPLICED or UNSPLICED */
bool
Equiv_class_revise (T this, unsigned int *transcript_indices, unsigned int *transcript_velocities,
		    Bitvector_T *retained_introns, Bitvector_T *spliced_introns, int ntranscripts) {
  Uintlist_T intersection_indices = NULL, intersection_velocities = NULL;
  List_T intersection_retained_introns = NULL, intersection_spliced_introns = NULL, p;
  Bitvector_T introns;

  unsigned int old_index, new_index, old_velocity, new_velocity;
  int *order;
  int i, j;

  order = Sedgesort_order(transcript_indices,ntranscripts);

  i = j = 0;
  while (i < this->ntranscripts && j < ntranscripts) {
    old_index = this->transcripts[i];
    new_index = transcript_indices[order[j]];
    if (old_index < new_index) {
      i++;

    } else if (new_index < old_index) {
      j++;

    } else if (Bitvector_intersect_p(retained_introns[order[j]],this->spliced_introns[i]) == true ||
	       Bitvector_intersect_p(spliced_introns[order[j]],this->retained_introns[i]) == true) {
      /* Conflict between unspliced and spliced introns for some transcript */
      Uintlist_free(&intersection_indices);
      Uintlist_free(&intersection_velocities);

      for (p = intersection_retained_introns; p != NULL; p = List_next(p)) {
	introns = (Bitvector_T) List_head(p);
	Bitvector_free(&introns);
      }
      List_free(&intersection_retained_introns);

      for (p = intersection_spliced_introns; p != NULL; p = List_next(p)) {
	introns = (Bitvector_T) List_head(p);
	Bitvector_free(&introns);
      }
      List_free(&intersection_spliced_introns);

      return false;

    } else {
      intersection_indices = Uintlist_push(intersection_indices,old_index);

      old_velocity = this->velocities[i];
      new_velocity = transcript_velocities[order[j]];
      if (old_velocity > new_velocity) {
	intersection_velocities = Uintlist_push(intersection_velocities,old_velocity);
      } else {
	intersection_velocities = Uintlist_push(intersection_velocities,new_velocity);
      }

      intersection_retained_introns = List_push(intersection_retained_introns,
						(void *) Bitvector_union(this->retained_introns[i],
									 retained_introns[order[j]],/*copyp*/true));
      intersection_spliced_introns = List_push(intersection_spliced_introns,
					       (void *) Bitvector_union(this->spliced_introns[i],
									spliced_introns[order[j]],/*copyp*/true));
      i++; j++;
    }
  }

  FREE(order);
  
  if (intersection_indices == NULL) {
    /* No transcripts were found in common */
    /* Three ways to handle: (1) Spin off this read as a new umi_id,
       (2) take the union of the transcripts, or (3) treat this UMI
       barcode as invalid (by setting this umi_id to -1).  If we spin
       off this read, then we mignt want to clean up UMI collsions
       based on their number of reads.  We are choosing option (1).
       We also need to resolve the conflicts, by picking the best
       equiv class among UMI collisions, and we designate losers by
       setting their umi_id to -1. */
    return false;

  } else {
    FREE(this->transcripts);
    FREE(this->velocities);
    for (i = 0; i < this->ntranscripts; i++) {
      Bitvector_free(&(this->retained_introns[i]));
      Bitvector_free(&(this->spliced_introns[i]));
    }
    FREE(this->retained_introns);
    FREE(this->spliced_introns);

    intersection_indices = Uintlist_reverse(intersection_indices);
    intersection_velocities = Uintlist_reverse(intersection_velocities);
    intersection_retained_introns = List_reverse(intersection_retained_introns);
    intersection_spliced_introns = List_reverse(intersection_spliced_introns);

    this->transcripts = Uintlist_to_array(&this->ntranscripts,intersection_indices);
    this->velocities = Uintlist_to_array(&this->ntranscripts,intersection_velocities);
    this->retained_introns = (Bitvector_T *) List_to_array(intersection_retained_introns,NULL);
    this->spliced_introns = (Bitvector_T *) List_to_array(intersection_spliced_introns,NULL);

    Uintlist_free(&intersection_indices);
    Uintlist_free(&intersection_velocities);
    List_free(&intersection_retained_introns);
    List_free(&intersection_spliced_introns);

    this->nreads += 1;

    return true;
  }
}


bool
Equiv_class_merge (T dest, T source) {
  Uintlist_T intersection_indices = NULL, intersection_velocities = NULL;
  List_T intersection_retained_introns = NULL, intersection_spliced_introns = NULL, p;
  Bitvector_T introns;

  unsigned int old_index, new_index, old_velocity, new_velocity;
  int i, j;

  i = j = 0;
  while (i < dest->ntranscripts && j < source->ntranscripts) {
    old_index = dest->transcripts[i];
    new_index = source->transcripts[j];
    if (old_index < new_index) {
      i++;

    } else if (new_index < old_index) {
      j++;

    } else if (Bitvector_intersect_p(source->retained_introns[j],dest->spliced_introns[i]) == true ||
	       Bitvector_intersect_p(source->spliced_introns[j],dest->retained_introns[i]) == true) {
      /* Conflict between unspliced and spliced introns for some transcript */
      Uintlist_free(&intersection_indices);
      Uintlist_free(&intersection_velocities);

      for (p = intersection_retained_introns; p != NULL; p = List_next(p)) {
	introns = (Bitvector_T) List_head(p);
	Bitvector_free(&introns);
      }
      List_free(&intersection_retained_introns);

      for (p = intersection_spliced_introns; p != NULL; p = List_next(p)) {
	introns = (Bitvector_T) List_head(p);
	Bitvector_free(&introns);
      }
      List_free(&intersection_spliced_introns);

      return false;

    } else {
      intersection_indices = Uintlist_push(intersection_indices,old_index);
      old_velocity = dest->velocities[i];
      new_velocity = source->velocities[j];
      if (old_velocity > new_velocity) {
	intersection_velocities = Uintlist_push(intersection_velocities,old_velocity);
      } else {
	intersection_velocities = Uintlist_push(intersection_velocities,new_velocity);
      }

      intersection_retained_introns = List_push(intersection_retained_introns,
						(void *) Bitvector_union(dest->retained_introns[i],
									 source->retained_introns[j],/*copyp*/true));
      intersection_spliced_introns = List_push(intersection_spliced_introns,
					       (void *) Bitvector_union(dest->spliced_introns[i],
									source->spliced_introns[j],/*copyp*/true));
      i++; j++;
    }
  }

  if (intersection_indices == NULL) {
    return false;

  } else {
    FREE(dest->transcripts);
    FREE(dest->velocities);
    for (i = 0; i < dest->ntranscripts; i++) {
      Bitvector_free(&(dest->retained_introns[i]));
      Bitvector_free(&(dest->spliced_introns[i]));
    }
    FREE(dest->retained_introns);
    FREE(dest->spliced_introns);

    intersection_indices = Uintlist_reverse(intersection_indices);
    intersection_velocities = Uintlist_reverse(intersection_velocities);
    intersection_retained_introns = List_reverse(intersection_retained_introns);
    intersection_spliced_introns = List_reverse(intersection_spliced_introns);

    dest->transcripts = Uintlist_to_array(&dest->ntranscripts,intersection_indices);
    dest->velocities = Uintlist_to_array(&dest->ntranscripts,intersection_velocities);
    dest->retained_introns = (Bitvector_T *) List_to_array(intersection_retained_introns,NULL);
    dest->spliced_introns = (Bitvector_T *) List_to_array(intersection_spliced_introns,NULL);

    Uintlist_free(&intersection_indices);
    Uintlist_free(&intersection_velocities);
    List_free(&intersection_retained_introns);
    List_free(&intersection_spliced_introns);

    dest->nreads += source->nreads;

    return true;
  }
}


int
Equiv_class_UMI_ndiffs (T x, T y) {
  Genomecomp_T diff;

  diff = (x->high ^ y->high) | (x->low ^ y->low);

  /* Consider N as a mismatch */
  diff |= x->flags;
  diff |= y->flags;

#if defined(HAVE_POPCNT)
  return _popcnt32(diff);
#elif defined(HAVE_MM_POPCNT)
  return _mm_popcnt_u32(diff);
#elif defined(HAVE_BUILTIN_POPCOUNT)
  return __builtin_popcount(diff);
#else
  return count_bits[diff & 0x0000FFFF] + count_bits[diff >> 16]);
#endif
}


static int
uint_cmp (const void *x, const void *y) {
  unsigned int a = * (int *) x;
  unsigned int b = * (int *) y;

  if (a < b) {
    return -1;
  } else if (b < a) {
    return +1;
  } else {
    return 0;
  }
}


/* Specific to single cell */
void
Equiv_class_set_genes (T this, unsigned int *transcripti_to_genei) {
  Uintlist_T gene_list = NULL;
  unsigned int *gene_array, genei;
  int ngenes, i, k;

  for (i = 0; i < this->ntranscripts; i++) {
    genei = transcripti_to_genei[this->transcripts[i]];
    if (gene_list == NULL) {
      gene_list = Uintlist_push(gene_list,genei);
    } else if (genei != Uintlist_head(gene_list)) {
      /* Adjacent transcript indices tend to have the same gene index, so can just check Uintlist_head */
      gene_list = Uintlist_push(gene_list,genei);
    } else {
      /* Already have this genei */
    }
  }

  gene_array = Uintlist_to_array(&ngenes,gene_list);
  qsort(gene_array,ngenes,sizeof(unsigned int),uint_cmp);
  Uintlist_free(&gene_list);

  this->ngenes = 1;
  for (i = 1; i < ngenes; i++) {
    if (gene_array[i] != gene_array[i-1]) {
      this->ngenes++;
    }
  }
  
  this->genes = (unsigned int *) MALLOC(this->ngenes*sizeof(unsigned int));
  this->genes[0] = gene_array[0];
  for (i = 1, k = 1; i < ngenes; i++) {
    if (gene_array[i] != gene_array[i-1]) {
      this->genes[k++] = gene_array[i];
    }
  }
    
  FREE(gene_array);

  return;
}


bool
Equiv_class_gene_intersection_p (T equiv_class_i, T equiv_class_j) {
  int i, j;

  i = j = 0;
  while (i < equiv_class_i->ngenes && j < equiv_class_j->ngenes) {
    if (equiv_class_i->genes[i] < equiv_class_j->genes[j]) {
      i++;
    } else if (equiv_class_j->genes[j] < equiv_class_i->genes[i]) {
      j++;
    } else {
      return true;
    }
  }
  
  return false;
}  


void
Equiv_class_add_introns (List_T *cell_transcript_introns, T this) {
  int i;
  unsigned int transcript_index;
  Bitvector_T introns;

  for (i = 0; i < this->ntranscripts; i++) {
    transcript_index = this->transcripts[i];
    if ((introns = this->retained_introns[i]) != NULL) {
      if (Bitvector_find(introns,cell_transcript_introns[transcript_index]) == false) {
	cell_transcript_introns[transcript_index] = List_push(cell_transcript_introns[transcript_index],(void *) Bitvector_copy(introns));
      }
    }
  }

  return;
}


void
Equiv_class_print_transcripts (FILE *out, T this, char *header,
			       unsigned int *transcripti_to_genei,char **all_genes) {
  int i;

  fprintf(out,"%s",header);
  for (i = 0; i < this->ntranscripts; i++) {
    fprintf(out," %d%c (%s)",
	   this->transcripts[i],velocity_char[this->velocities[i]],
	   all_genes[transcripti_to_genei[this->transcripts[i]]]);
  }
  fprintf(out,"\n");

  return;
}

int
Equiv_class_nreads (T this) {
  return this->nreads;
}

void
Equiv_class_combine_nreads (T dest, T source) {
  dest->nreads += source->nreads;
  source->nreads = 0;
  return;
}

void
Equiv_class_print (T this) {
  int i;

  for (i = 0; i < this->ntranscripts; i++) {
    printf(" %u",this->transcripts[i]);
    if (this->velocities[i] == SPLICED) {
      printf("S");
    } else if (this->velocities[i] == UNSPLICED) {
      printf("U");
    } else if (this->velocities[i] == BOTH) {
      printf("B");
    }
  }
  printf("\n");
}

/* Specific to single cell */
/* Equiv_class_nreads_cmp is called on pointers to structures by qsort */
int
Equiv_class_nreads_cmp (const void *a, const void *b) {
  T x = * (T *) a;
  T y = * (T *) b;

  if (x->nreads > y->nreads) {
    return -1;
  } else if (y->nreads > x->nreads) {
    return +1;
  } else {
    return 0;
  }
}


void
Equiv_class_calc_weights (T this, double *spliced_lengths, Doublelist_T *unspliced_lengths,
			  List_T *cell_transcript_introns) {
  int i;
  unsigned int transcript_index;
  unsigned int velocity_type;
  List_T p;
  Doublelist_T q;
  Bitvector_T introns;
  double unspliced_length;

  if (this->ntranscripts > 0) {
    this->spliced_weights = (double *) CALLOC(this->ntranscripts,sizeof(double));
    this->unspliced_weights = (Doublelist_T *) CALLOC(this->ntranscripts,sizeof(Doublelist_T));

    for (i = 0; i < this->ntranscripts; i++) {
      transcript_index = this->transcripts[i];
      
      if ((introns = this->retained_introns[i]) != NULL) {
	/* Match one unspliced form */
	this->spliced_weights[i] = 0.0;
	unspliced_length = Bitvector_double_find(this->retained_introns[i],cell_transcript_introns[transcript_index],
						 unspliced_lengths[transcript_index]);
	this->unspliced_weights[i] = Doublelist_push(NULL,/*count*/1.0 / unspliced_length);

      } else if ((introns = this->spliced_introns[i]) != NULL) {
	/* Match spliced plus compatible unspliced forms */
	this->spliced_weights[i] = /*count*/1.0 / spliced_lengths[transcript_index];
	this->unspliced_weights[i] = (Doublelist_T) NULL;
	for (p = cell_transcript_introns[transcript_index], q = unspliced_lengths[transcript_index]; q != NULL; q = Doublelist_next(q)) {
	  if (Bitvector_intersect_p(introns,(Bitvector_T) List_head(p)) == false) {
	    unspliced_length = Doublelist_head(q);
	    this->unspliced_weights[i] = Doublelist_push(this->unspliced_weights[i],/*count*/1.0 / unspliced_length);
	  }
	}
	this->unspliced_weights[i] = Doublelist_reverse(this->unspliced_weights[i]);

      } else {
	/* Match spliced plus all unspliced forms */
	this->spliced_weights[i] = /*count*/1.0 / spliced_lengths[transcript_index];
	this->unspliced_weights[i] = (Doublelist_T) NULL;
	for (q = unspliced_lengths[transcript_index]; q != NULL; q = Doublelist_next(q)) {
	  unspliced_length = Doublelist_head(q);
	  this->unspliced_weights[i] = Doublelist_push(this->unspliced_weights[i],/*count*/1.0 / unspliced_length);
	}
	this->unspliced_weights[i] = Doublelist_reverse(this->unspliced_weights[i]);
      }
    }
  }

  return;
}


#define TOLERANCE 2.22045e-15

/* UNSPLICED */
/* If unspliced introns are 4, then compatible just with unspliced 4 */
/* We can ignore spliced_introns */
/* Not compatible with fully spliced */
/* So include in alpha_unspliced as long as Bitvector_equal_p(this->spliced_introns,cell_transcript_introns[transcript_index]) == true, or
   Bitvector_diff_p(this->spliced_introns,cell_transcript_introns[transcript_index]) == false */

/* SPLICED */
/* Else if unspliced introns is NULL, then look at spliced_introns */
/* Compatible with fully spliced */
/* If spliced introns are 1,2,3, then also compatible with unspliced 4 */
/* So include in alpha_unspliced as long as Bitvector_intersect_p(this->spliced_introns,cell_transcript_introns[transcript_index]) == false */

/* BOTH */
/* Else if both unspliced_introns is NULL and spliced_introns is NULL, then compatible with fully spliced and all unspliced */


void
Equiv_class_EM_update (double *next_alpha_spliced, Doublelist_T *next_alpha_unspliced, T this,
		       double *alpha_spliced, Doublelist_T *alpha_unspliced, List_T *cell_transcript_introns) {
  double denom = 0.0, count_norm, alpha;
  unsigned int transcript_index;
  unsigned int velocity_type;
  List_T p;
  Doublelist_T q, nextq, r, ptr;
  Bitvector_T introns;
  int i;

#if 1
  for (i = 0; i < this->ntranscripts; i++) {
    transcript_index = this->transcripts[i];
    
    if ((introns = this->retained_introns[i]) != NULL) {
      /* Match one unspliced form */
      alpha = Bitvector_double_find(introns,cell_transcript_introns[transcript_index],alpha_unspliced[transcript_index]);
      denom += alpha * Doublelist_head(this->unspliced_weights[i]);

    } else if ((introns = this->spliced_introns[i]) != NULL) {
      /* Match spliced plus compatible unspliced forms */
      denom += alpha_spliced[transcript_index] * this->spliced_weights[i];
      for (p = cell_transcript_introns[transcript_index], q = alpha_unspliced[transcript_index], r = this->unspliced_weights[i];
	   p != NULL; p = List_next(p), q = Doublelist_next(q)) {
	if (Bitvector_intersect_p(introns,(Bitvector_T) List_head(p)) == false) {
	  denom += Doublelist_head(q) * Doublelist_head(r);
	  r = Doublelist_next(r);
	}
      }

    } else {
      /* Match spliced plus all unspliced forms */
      denom += alpha_spliced[transcript_index] * this->spliced_weights[i];
      for (q = alpha_unspliced[transcript_index], r = this->unspliced_weights[i];
	   q != NULL; q = Doublelist_next(q), r = Doublelist_next(r)) {
	denom += Doublelist_head(q) * Doublelist_head(r);
      }
    }
  }

  if (denom >= TOLERANCE) {
    count_norm = /*count*/1.0 / denom;
    for (i = 0; i < this->ntranscripts; i++) {
      transcript_index = this->transcripts[i];

      if ((introns = this->retained_introns[i]) != NULL) {
	/* Match one unspliced form */
	alpha = Bitvector_double_find(introns,cell_transcript_introns[transcript_index],alpha_unspliced[transcript_index]);
	ptr = Bitvector_double_find_ptr(introns,cell_transcript_introns[transcript_index],next_alpha_unspliced[transcript_index]);
	Doublelist_head_add(ptr,alpha * Doublelist_head(this->unspliced_weights[i]) * count_norm);

      } else if ((introns = this->spliced_introns[i]) != NULL) {
	/* Match spliced plus compatible unspliced forms */
	next_alpha_spliced[transcript_index] +=
	  alpha_spliced[transcript_index] * this->spliced_weights[i] * count_norm;

	for (p = cell_transcript_introns[transcript_index], q = alpha_unspliced[transcript_index], nextq = next_alpha_unspliced[transcript_index], r = this->unspliced_weights[i]; p != NULL;
	     p = List_next(p), q = Doublelist_next(q), nextq = Doublelist_next(nextq)) {
	  if (Bitvector_intersect_p(introns,(Bitvector_T) List_head(p)) == false) {
	    Doublelist_head_add(nextq,Doublelist_head(q) * Doublelist_head(r) * count_norm);
	    r = Doublelist_next(r);
	  }
	}

      } else {
	/* Match spliced plus all unspliced forms */
	next_alpha_spliced[transcript_index] +=
	  alpha_spliced[transcript_index] * this->spliced_weights[i] * count_norm;

	for (q = alpha_unspliced[transcript_index], nextq = next_alpha_unspliced[transcript_index],
	       r = this->unspliced_weights[i]; q != NULL;
	     q = Doublelist_next(q), nextq = Doublelist_next(nextq), r = Doublelist_next(r)) {
	  Doublelist_head_add(nextq,Doublelist_head(q) * Doublelist_head(r) * count_norm);
	}
      }
    }
  }

#else
  for (i = 0; i < this->ntranscripts; i++) {
    transcript_index = this->transcripts[i];

    if ((velocity_type = this->velocities[i]) == SPLICED) {
      denom += alpha_spliced[transcript_index] * this->spliced_weights[i];
      
    } else if (velocity_type == UNSPLICED) {
      alpha = Bitvector_double_find(this->retained_introns[i],all_transcript_introns[transcript_index],
				    alpha_unspliced[transcript_index]);
      denom += alpha * Doublelist_head(this->unspliced_weights[i]);
      
    } else if (velocity_type == BOTH) {
      denom += alpha_spliced[transcript_index] * this->spliced_weights[i];
      for (p = alpha_unspliced[transcript_index], q = this->unspliced_weights[i];
	   p != NULL; p = Doublelist_next(p), q = Doublelist_next(q)) {
	denom += Doublelist_head(p) * Doublelist_head(q);
      }
    }
  }

  if (denom >= TOLERANCE) {
    count_norm = /*count*/1.0 / denom;
    for (i = 0; i < this->ntranscripts; i++) {
      transcript_index = this->transcripts[i];
      if ((velocity_type = this->velocities[i]) == SPLICED) {
	next_alpha_spliced[transcript_index] +=
	  alpha_spliced[transcript_index] * this->spliced_weights[i] * count_norm;
	
      } else if (velocity_type == UNSPLICED) {
	alpha = Bitvector_double_find(this->introns[i],all_transcript_introns[transcript_index],alpha_unspliced[transcript_index]);
	ptr = Bitvector_double_find_ptr(this->introns[i],all_transcript_introns[transcript_index],next_alpha_unspliced[transcript_index]);
	Doublelist_head_add(ptr,alpha * Doublelist_head(this->unspliced_weights[i]) * count_norm);
	
      } else if (velocity_type == BOTH) {
	next_alpha_spliced[transcript_index] +=
	  alpha_spliced[transcript_index] * this->spliced_weights[i] * count_norm;
	
	for (p = alpha_unspliced[transcript_index], nextp = next_alpha_unspliced[transcript_index],
	       q = this->unspliced_weights[i]; p != NULL;
	     p = Doublelist_next(p), nextp = Doublelist_next(nextp), q = Doublelist_next(q)) {
	  Doublelist_head_add(nextp,Doublelist_head(p) * Doublelist_head(q) * count_norm);
	}
      }
    }
  }
#endif



  return;
}

