static char rcsid[] = "$Id$";
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>		/* For isdigit */

#if defined(HAVE_POPCNT)
#include <immintrin.h>
#elif defined(HAVE_MM_POPCNT)
#include <nmmintrin.h>
#endif

#include "assert.h"
#include "bool.h"
#include "types.h"
#include "iit-read.h"

#include "getline.h"
#include "samread.h"
#include "intlist.h"
#include "list.h"
#include "table.h"
#include "tableint.h"
#include "popcount.h"

#include "datadir.h"
#include "getopt.h"


#define nottab(x) (x) != '\t'

typedef enum {CELLS_AS_COLUMNS, CELLS_AS_ROWS} Matrix_order_T;


/* Need genomedir to get transcript_map_iit, which has the gene symbols */
static char *user_genomedir = NULL;
static char *genome_dbroot = NULL;
static char *transcriptome_dbroot = NULL;

static bool bulk_rna_p = false;
static char *input_command = NULL;
static int paired_end_splicedist = 1000;
static char *output_prefix = NULL;
static char *prior_file = NULL;
static char *correction_file = NULL;
static Matrix_order_T matrix_order = CELLS_AS_COLUMNS;
static int min_umis = 1000;

static char *bulk_cell = "NA";
static char *bulk_cell_name_ptr;
static char *bulk_umi_barcode_ptrs[4];


#ifdef DEBUG
#define debug(x) x
#else
#define debug(x)
#endif



static struct option long_options[] = {
  /* Input options */
  {"dir", required_argument, 0, 'D'},	/* user_genomedir */
  {"db", required_argument, 0, 'd'}, /* genome_dbroot */
  {"transcriptdir", required_argument, 0, 'C'},	/* user_transcriptomedir */
  {"transcriptdb", required_argument, 0, 'c'}, /* transcriptome_dbroot */
  {"input-command", required_argument, 0, 0},		       /* input_command */
  {"paired-end-splicedist", required_argument, 0, 0},	       /* paired_end_splicedist */
  {"bulk", no_argument, 0, 0},				       /* bulk_rna_p */

  /* Output options */
  {"output", required_argument, 0, 'o'}, /* output_prefix */
  {"priors", required_argument, 0, 0}, /* prior_file */
  {"umi-corrections", required_argument, 0, 0}, /* correction_file */
  {"matrix-columms", required_argument, 0, 0}, /* matrix_order */
  {"min-umis", required_argument, 0, 0},       /* min_umis */

  {"version", no_argument, 0, 0}, /* print_program_version */
  {"help", no_argument, 0, 0}, /* print_program_usage */
  {0, 0, 0, 0}
};

static void
print_program_version () {
  char *genomedir;

  fprintf(stdout,"\n");
  fprintf(stdout,"velocity-counts: Parses GSNAP output from transcriptome-guided genomic alignment\n");
  fprintf(stdout,"Part of GMAP package, version %s\n",PACKAGE_VERSION);

  fprintf(stdout,"Default gmap directory (compiled): %s\n",GMAPDB);
  genomedir = Datadir_find_genomedir(/*user_genomedir*/NULL);
  fprintf(stdout,"Default gmap directory (environment): %s\n",genomedir);
  FREE(genomedir);

  fprintf(stdout,"Thomas D. Wu, Genentech, Inc.\n");
  fprintf(stdout,"Contact: twu@gene.com\n");
  fprintf(stdout,"\n");

  return;
}

static void
print_program_usage ();


/* transcript_dict: transcript_name -> transcript_index */
static Tableint_T
get_transcripts (int *n_total_transcripts, IIT_T transcript_map_iit) {
  Tableint_T transcript_dict;
  char *transcript_name, *transcript_name_ptr;
  int indx;
  bool allocp;

  *n_total_transcripts = IIT_total_nintervals(transcript_map_iit);

  transcript_dict = Tableint_new(/*hint*/*n_total_transcripts,
                                  Table_string_compare,Table_string_hash);

  for (indx = 1; indx <= *n_total_transcripts; indx++) {
    transcript_name_ptr = IIT_label(transcript_map_iit,indx,&allocp);
    transcript_name = (char *) MALLOC((strlen(transcript_name_ptr)+1)*sizeof(char));
    strcpy(transcript_name,transcript_name_ptr);
    Tableint_put(transcript_dict,(void *) transcript_name,indx); /* Store as 1-based */
  }

  return transcript_dict;
}


static char *
transcript_gene_name (int *stringlen, char *annot) {
  char *gene, *p;

  /* Parse header */
  p = annot;
  while (*p != '\0' && *p != '\n' && !isspace(*p)) {
    p++;
  }
  *stringlen = p - annot;

  gene = (char *) MALLOC(((*stringlen)+1)*sizeof(char));
  strncpy(gene,annot,*stringlen);
  gene[*stringlen] = '\0';

  return gene;
}


/* transcripti_to_genei: transcript_index -> gene_index */
static char **
get_genes (int *n_total_genes, int **transcripti_to_genei, Intlist_T **genei_to_transcripti,
	   IIT_T transcript_map_iit) {
  char **all_genes;
  List_T gene_list = NULL;
  Intlist_T index_list = NULL;

  int gene_index, ngenes = 0, n_total_transcripts,
    transcript_index, indx;
  int gene_namelen;
  int ignore;
  Tableint_T gene_table;
  char *annot, *restofheader, *symbol;
  bool allocp;

  n_total_transcripts = IIT_total_nintervals(transcript_map_iit);
  gene_table = Tableint_new(/*hint*/n_total_transcripts,
                            Table_string_compare,Table_string_hash);

  for (indx = 1; indx <= n_total_transcripts; indx++) {
    annot = IIT_annotation(&restofheader,transcript_map_iit,indx,&allocp);
    symbol = (char *) transcript_gene_name(&gene_namelen,annot);
    /* gene_table stores gene-indices as 1-based, so 0 means not found */
    if ((gene_index = Tableint_get(gene_table,(void *) symbol)) == 0) {
      Tableint_put(gene_table,(void *) symbol,++ngenes);
      gene_list = List_push(gene_list,(void *) symbol);
      index_list = Intlist_push(index_list,ngenes-1); /* 0-based */
    } else {
      index_list = Intlist_push(index_list,gene_index-1); /* 0-based */
      FREE(symbol);
    }
    if (allocp) {
      FREE(restofheader);
    }
  }

  *n_total_genes = ngenes;
  gene_list = List_reverse(gene_list);
  all_genes = (char **) List_to_array(gene_list,NULL);
  List_free(&gene_list);

  index_list = Intlist_reverse(index_list);
  *transcripti_to_genei = Intlist_to_array(&ignore,index_list); /* 0-based */
  Intlist_free(&index_list);

  *genei_to_transcripti = (Intlist_T *) CALLOC(ngenes,sizeof(Intlist_T));
  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    gene_index = (*transcripti_to_genei)[transcript_index]; /* 0-based */
    index_list = (*genei_to_transcripti)[gene_index];
    (*genei_to_transcripti)[gene_index] = Intlist_push(index_list,transcript_index); /* 0-based */
  }
  
  Tableint_free(&gene_table);

  return all_genes;
}


/*                    87654321 */
#define RIGHT_SET   0x00000001
#define RIGHT_CLEAR 0x00000000


/* Genome128 format */
/*          High bit   Low bit   Flag bit
   A = 000     0          0         0
   C = 001     0          1         0
   G = 010     1          0         0
   T = 011     1          1         1
   N = 100     0          0         1
   X = 111     1          1         1
*/

static void
compress_umi (Genomecomp_T *high, Genomecomp_T *low, Genomecomp_T *flags,
	      char *umi) {
  int pos, length;
  char c;

  length = strlen(umi);
  assert(length <= 32);
  
  *high = *low = *flags = 0U;
  for (pos = 0; pos < length; pos++) {
    c = umi[pos];
    *high <<= 1;
    *low <<= 1;
    *flags <<= 1;

    /* Assume that umi is upper case */
    switch /*(uppercaseCode[c])*/ (c) {
    case 'A': /* *high |= RIGHT_CLEAR; */ /* *low |= RIGHT_CLEAR; */ /* *flags |= RIGHT_CLEAR; */ break;
    case 'C': /* *high |= RIGHT_CLEAR; */    *low |= RIGHT_SET;      /* *flags |= RIGHT_CLEAR; */ break;
    case 'G':    *high |= RIGHT_SET;      /* *low |= RIGHT_CLEAR; */ /* *flags |= RIGHT_CLEAR; */ break;
    case 'T':    *high |= RIGHT_SET;         *low |= RIGHT_SET;      /* *flags |= RIGHT_CLEAR; */ break;
    default:  /* *high |= RIGHT_CLEAR; */ /* *low |= RIGHT_CLEAR; */    *flags |= RIGHT_SET;
    }
  }

  return;
}


/* Unspliced takes precedence of spliced, which takes precedence over ambiguous */
typedef enum {UNSEEN, AMBIGUOUS, SPLICED, UNSPLICED} Velocity_T;

typedef struct UMI_T *UMI_T;
struct UMI_T {
  char *barcode_ptr;
  Genomecomp_T high, low, flags;
  int nreads;
  Velocity_T velocity_type;
};


static UMI_T
UMI_new (char *barcode_ptr, Velocity_T velocity_type) {
  UMI_T new = (UMI_T) MALLOC(sizeof(*new));

  new->barcode_ptr = barcode_ptr;
  compress_umi(&new->high,&new->low,&new->flags,barcode_ptr);
  new->nreads = 1;
  new->velocity_type = velocity_type;

  return new;
}

static void
UMI_free (UMI_T *old) {
  FREE(*old);
  return;
}

static int
UMI_cmp (const void *a, const void *b) {
  UMI_T x = * (UMI_T *) a;
  UMI_T y = * (UMI_T *) b;

  if (x->nreads > y->nreads) {
    return -1;
  } else if (y->nreads > x->nreads) {
    return +1;
  } else {
    return 0;
  }
}



static int
UMI_ndiffs (UMI_T x, UMI_T y) {
  Genomecomp_T diff;

  diff = (x->high ^ y->high) | (x->low ^ y->low);

  /* Consider N as a mismatch */
  diff |= x->flags;
  diff |= y->flags;

#if defined(HAVE_POPCNT)
  return _popcnt32(diff);
#elif defined(HAVE_MM_POPCNT)
  return _mm_popcnt_u32(diff);
#elif defined(HAVE_BUILTIN_POPCOUNT)
  return __builtin_popcount(diff);
#else
  return count_bits[diff & 0x0000FFFF] + count_bits[diff >> 16]);
#endif
}


static void
update_umi_list (List_T *umi_list, char *umi_barcode_ptr, Velocity_T velocity_type) {
  List_T p = *umi_list;
  UMI_T umi;

  while (p != NULL) {
    umi = (UMI_T) List_head(p);
    if (umi->barcode_ptr == umi_barcode_ptr) {
      umi->nreads += 1;
      if (velocity_type > umi->velocity_type) {
	umi->velocity_type = velocity_type;
      }
      return;

    } else {
      p = List_next(p);
    }
  }

  *umi_list = List_push(*umi_list,(void *) UMI_new(umi_barcode_ptr,velocity_type));

  return;
}


/* For single-cell analysis */
static double *
collect_prior_and_umis_block (double *prior, List_T lines, Table_T *cell_umi_tables,
			      Tableint_T transcript_dict, Table_T cell_dict, Table_T umi_dict) {
  Table_T cell_umi_table;
  List_T umi_list;
  Tableint_T transcript_seenp;

  double incr;
  int ntranscripts, i;
  char **transcripts, *transcript_list, *transcript_name, velocity_char;
  Velocity_T velocity_type;
  int transcript_index;
  char *line, *auxinfo, *p, *q;

  char *cell_name_ptr, *cell, *umi_barcode_ptr, *umi;

  unsigned int flag;
  int insertlength;
  int namelen;

  List_T a;


  /* We assume that the CB and UB/UR fields are the same for the entire block */
  line = (char *) List_head(lines);
  auxinfo = Samread_auxinfo(&flag,&insertlength,line);

  if ((cell = Samread_get_aux_string("CB",auxinfo)) == NULL) {
    /* Skip block that has no cell */
    return prior;
    
  } else if ((umi = Samread_get_aux_string("UB",auxinfo)) == NULL &&
	     (umi = Samread_get_aux_string("UR",auxinfo)) == NULL) {
    FREE(cell);

  } else {
    if ((cell_name_ptr = Table_get(cell_dict,(void *) cell)) != NULL) {
      FREE(cell);
    } else {
      cell_name_ptr = cell;
      Table_put(cell_dict,(void *) cell,(void *) cell_name_ptr);
    }

    if ((umi_barcode_ptr = Table_get(umi_dict,(void *) umi)) != NULL) {
      FREE(umi);
    } else {
      umi_barcode_ptr = umi;
      Table_put(umi_dict,(void *) umi,(void *) umi_barcode_ptr);
    }
  }

  /* Count valid transcripts */
  transcript_seenp = Tableint_new(/*hint*/10,Table_string_compare,Table_string_hash);
  ntranscripts = 0;
  for (a = lines; a != NULL; a = List_next(a)) {
    line = (char *) List_head(a);
    auxinfo = Samread_auxinfo(&flag,&insertlength,line);

    if ((transcript_list = Samread_get_aux_string("XX",auxinfo)) != NULL) {
      p = transcript_list;
      while (*p != '\0') {
	p++;			/* Skip [+-] */
	q = p;
	while (*q++ != ':') {
	  /* Advance through transcript name */
	}
	namelen = q - p - 1;
	while (*q++ != ':') {
	  /* Advance through transcript coordinates */
	}

	if ((velocity_char = *q) == 'X' || velocity_char == 'E') {
	  /* Skip */
	} else {
	  if (velocity_char == 'U') {
	    velocity_type = UNSPLICED;
	  } else if (velocity_char == 'S') {
	    velocity_type = SPLICED;
	  } else if (insertlength > paired_end_splicedist || insertlength < -paired_end_splicedist) {
	    velocity_type = SPLICED;
	  } else {
	    velocity_type = AMBIGUOUS;
	  }

	  transcript_name = (char *) MALLOC((namelen+1)*sizeof(char));
	  strncpy(transcript_name,p,namelen);
	  transcript_name[namelen] = '\0';

	  if (Tableint_get(transcript_seenp,(void *) transcript_name) > UNSEEN) {
	    FREE(transcript_name);
	  } else {
	    Tableint_put(transcript_seenp,(void *) transcript_name,(int) velocity_type);
	    ntranscripts++;
	  }
	}

	while (*p != '\0' && *p != ',') {
	  p++;
	}
	if (*p == ',') {
	  p++;
	}
      }
      FREE(transcript_list);
    }
  }

  /* Enter priors and UMIs */
  if (ntranscripts > 0) {
    /* incr = 1.0/(double) ntranscripts; */
    incr = 1.0;
	
    transcripts = (char **) Tableint_keys(transcript_seenp,NULL);
    for (i = 0; i < ntranscripts; i++) {
      transcript_name = transcripts[i];
      velocity_type = (Velocity_T) Tableint_get(transcript_seenp,(void *) transcript_name);
      transcript_index = Tableint_get(transcript_dict,(void *) transcript_name) - 1; /* 0-based */
      prior[transcript_index] += incr;
	    
      /* printf("Storing into transcript %d %s, type %d, cell %s, umi %s\n",
	 transcript_index,transcript_name,velocity_type,cell_name_ptr,umi_barcode_ptr); */

      if ((cell_umi_table = cell_umi_tables[transcript_index]) == (Table_T) NULL) {
	cell_umi_table = cell_umi_tables[transcript_index] = Table_new(/*hint*/1000,NULL,NULL);
      }

      umi_list = (List_T) Table_get(cell_umi_table,(void *) cell_name_ptr);
      update_umi_list(&umi_list,umi_barcode_ptr,velocity_type);
      Table_put(cell_umi_table,(void *) cell_name_ptr,(void *) umi_list);

      FREE(transcript_name);
    }
    FREE(transcripts);
  }

  Tableint_free(&transcript_seenp);

  return prior;
}


/* For bulk RNA-seq analysis */
static double *
tally_prior_block (double *prior, List_T lines, Tableint_T transcript_dict) {
  Tableint_T transcript_seenp;

  double incr;
  int ntranscripts, i;
  char **transcripts, *transcript_list, *transcript_name, velocity_char;
  Velocity_T velocity_type;
  int transcript_index;
  char *line, *auxinfo, *p, *q;

  unsigned int flag;
  int insertlength;
  int namelen;

  List_T a;


  /* We assume that the CB and UB/UR fields are the same for the entire block */
  line = (char *) List_head(lines);
  auxinfo = Samread_auxinfo(&flag,&insertlength,line);

  /* Count valid transcripts */
  transcript_seenp = Tableint_new(/*hint*/10,Table_string_compare,Table_string_hash);
  ntranscripts = 0;
  for (a = lines; a != NULL; a = List_next(a)) {
    line = (char *) List_head(a);
    auxinfo = Samread_auxinfo(&flag,&insertlength,line);

    if ((transcript_list = Samread_get_aux_string("XX",auxinfo)) != NULL) {
      p = transcript_list;
      while (*p != '\0') {
	p++;			/* Skip [+-] */
	q = p;
	while (*q++ != ':') {
	  /* Advance through transcript name */
	}
	namelen = q - p - 1;
	while (*q++ != ':') {
	  /* Advance through transcript coordinates */
	}

	if ((velocity_char = *q) == 'X' || velocity_char == 'E') {
	  /* Skip */
	} else {
	  if (velocity_char == 'U') {
	    velocity_type = UNSPLICED;
	  } else if (velocity_char == 'S') {
	    velocity_type = SPLICED;
	  } else if (insertlength > paired_end_splicedist || insertlength < -paired_end_splicedist) {
	    velocity_type = SPLICED;
	  } else {
	    velocity_type = AMBIGUOUS;
	  }

	  transcript_name = (char *) MALLOC((namelen+1)*sizeof(char));
	  strncpy(transcript_name,p,namelen);
	  transcript_name[namelen] = '\0';

	  if (Tableint_get(transcript_seenp,(void *) transcript_name) > UNSEEN) {
	    FREE(transcript_name);
	  } else {
	    Tableint_put(transcript_seenp,(void *) transcript_name,(int) velocity_type);
	    ntranscripts++;
	  }
	}

	while (*p != '\0' && *p != ',') {
	  p++;
	}
	if (*p == ',') {
	  p++;
	}
      }
      FREE(transcript_list);
    }
  }

  /* Enter priors */
  if (ntranscripts > 0) {
    /* incr = 1.0/(double) ntranscripts; */
    incr = 1.0;
	
    transcripts = (char **) Tableint_keys(transcript_seenp,NULL);
    for (i = 0; i < ntranscripts; i++) {
      transcript_name = transcripts[i];
      transcript_index = Tableint_get(transcript_dict,(void *) transcript_name) - 1; /* 0-based */
      prior[transcript_index] += incr;
      FREE(transcript_name);
    }
    FREE(transcripts);
  }

  Tableint_free(&transcript_seenp);

  return prior;
}


/* For bulk RNA-seq analysis */
static void
collect_umis_block (List_T lines, double *prior, Table_T *cell_umi_tables,
		    Tableint_T transcript_dict) {
  Table_T cell_umi_table;
  List_T umi_list;

  double max_prior;
  char *transcript_list, *transcript_name, velocity_char;
  Velocity_T best_velocity_type, velocity_type;
  int best_transcript_index, transcript_index;
  char *line, *auxinfo, *p, *q;

  char *cell_name_ptr, *umi_barcode_ptr;

  unsigned int flag;
  int insertlength;
  int namelen;

  List_T a;


  /* We assume that the CB and UB/UR fields are the same for the entire block */
  line = (char *) List_head(lines);
  auxinfo = Samread_auxinfo(&flag,&insertlength,line);

  cell_name_ptr = bulk_cell_name_ptr;
  /* Determine "UMI barcode/name" below based on velocity_type */

  /* Find the transcript with the max prior */
  max_prior = 0.0;
  best_transcript_index = -1;
  best_velocity_type = UNSEEN;

  for (a = lines; a != NULL; a = List_next(a)) {
    line = (char *) List_head(a);
    auxinfo = Samread_auxinfo(&flag,&insertlength,line);

    if ((transcript_list = Samread_get_aux_string("XX",auxinfo)) != NULL) {
      p = transcript_list;
      while (*p != '\0') {
	p++;			/* Skip [+-] */
	q = p;
	while (*q++ != ':') {
	  /* Advance through transcript name */
	}
	namelen = q - p - 1;
	while (*q++ != ':') {
	  /* Advance through transcript coordinates */
	}

	if ((velocity_char = *q) == 'X' || velocity_char == 'E') {
	  /* Skip */
	} else {
	  if (velocity_char == 'U') {
	    velocity_type = UNSPLICED;
	  } else if (velocity_char == 'S') {
	    velocity_type = SPLICED;
	  } else if (insertlength > paired_end_splicedist || insertlength < -paired_end_splicedist) {
	    velocity_type = SPLICED;
	  } else {
	    velocity_type = AMBIGUOUS;
	  }

	  transcript_name = (char *) MALLOC((namelen+1)*sizeof(char));
	  strncpy(transcript_name,p,namelen);
	  transcript_name[namelen] = '\0';
	  transcript_index = Tableint_get(transcript_dict,(void *) transcript_name) - 1; /* 0-based */
	  FREE(transcript_name);

	  if (prior[transcript_index] > max_prior) {
	    max_prior = prior[transcript_index];
	    best_transcript_index = transcript_index;
	    best_velocity_type = velocity_type;
	  }
	}

	while (*p != '\0' && *p != ',') {
	  p++;
	}
	if (*p == ',') {
	  p++;
	}
      }

      FREE(transcript_list);
    }
  }

  /* Enter UMI */
  if (best_transcript_index >= 0) {
    umi_barcode_ptr = bulk_umi_barcode_ptrs[best_velocity_type];

    /* printf("Storing into transcript %d %s, type %d, cell %s, umi %s\n",
       best_transcript_index,transcript_name,velocity_type,cell_name_ptr,umi_barcode_ptr); */

    if ((cell_umi_table = cell_umi_tables[best_transcript_index]) == (Table_T) NULL) {
      cell_umi_table = cell_umi_tables[best_transcript_index] = Table_new(/*hint*/1000,NULL,NULL);
    }

    umi_list = (List_T) Table_get(cell_umi_table,(void *) cell_name_ptr);
    update_umi_list(&umi_list,umi_barcode_ptr,best_velocity_type);
    Table_put(cell_umi_table,(void *) cell_name_ptr,(void *) umi_list);
  }

  return;
}


static void
lines_gc (List_T *lines) {
  List_T p;
  char *line;

  for (p = *lines; p != NULL; p = List_next(p)) {
    line = (char *) List_head(p);
    FREE(line);
  }
  List_free(&(*lines));
  *lines = (List_T) NULL;

  return;
}


/* This needs to allocate the string */
static char *
get_acc (char *line) {
  char *acc, *p;
  int length;

  p = line;
  while (nottab(*p++)) ;
  length = (p - 1 - line)/sizeof(char);
  acc = (char *) MALLOC((length+1)*sizeof(char));
  strncpy(acc,line,length);
  acc[length] = '\0';

  return acc;
}


static double *
tally_prior (FILE *fp, Tableint_T transcript_dict) {
  double *prior;

  List_T lines;
  char *line, *acc, *lastacc;
  int linelen;
  int n_total_transcripts;


  n_total_transcripts = Tableint_length(transcript_dict);
  prior = (double *) CALLOC(n_total_transcripts,sizeof(double));

  lastacc = (char *) MALLOC(1*sizeof(char));
  lastacc[0] = '\0';
  lines = (List_T) NULL;
  while ((line = Getline_wlinefeed(&linelen,fp)) != NULL) {
    acc = get_acc(line);
    if (!strcmp(acc,lastacc)) {
      /* Same acc as before */
      FREE(acc);
      acc = lastacc;
    } else if (lines == NULL) {
      /* First acc seen */
      FREE(lastacc);
      lastacc = acc;
    } else {
      /* Process lines */
      prior = tally_prior_block(prior,lines,transcript_dict);
      lines_gc(&lines);
      FREE(lastacc);
      lastacc = acc;
    }
    lines = List_push(lines,(void *) line);
  }

  if (lines == NULL) {
    /* No lines seen */
    FREE(lastacc);
  } else {
    /* Process lines */
    prior = tally_prior_block(prior,lines,transcript_dict);
    lines_gc(&lines);
    FREE(lastacc);
  }

  return prior;
}


static void
collect_umis (FILE *fp, double *prior, Table_T *cell_umi_tables,
	      Tableint_T transcript_dict) {
  List_T lines;
  char *line, *acc, *lastacc;
  int linelen;


  lastacc = (char *) MALLOC(1*sizeof(char));
  lastacc[0] = '\0';
  lines = (List_T) NULL;
  while ((line = Getline_wlinefeed(&linelen,fp)) != NULL) {
    acc = get_acc(line);
    if (!strcmp(acc,lastacc)) {
      /* Same acc as before */
      FREE(acc);
      acc = lastacc;
    } else if (lines == NULL) {
      /* First acc seen */
      FREE(lastacc);
      lastacc = acc;
    } else {
      /* Process lines */
      collect_umis_block(lines,prior,cell_umi_tables,transcript_dict);
      lines_gc(&lines);
      FREE(lastacc);
      lastacc = acc;
    }
    lines = List_push(lines,(void *) line);
  }

  if (lines == NULL) {
    /* No lines seen */
    FREE(lastacc);
  } else {
    /* Process lines */
    collect_umis_block(lines,prior,cell_umi_tables,transcript_dict);
    lines_gc(&lines);
    FREE(lastacc);
  }

  return;
}


static double *
collect_prior_and_umis (FILE *fp, Table_T *cell_umi_tables,
			Tableint_T transcript_dict, Table_T cell_dict, Table_T umi_dict) {
  double *prior;
  List_T lines;
  char *line, *acc, *lastacc;
  int linelen;
  int n_total_transcripts;

  n_total_transcripts = Tableint_length(transcript_dict);
  prior = (double *) CALLOC(n_total_transcripts,sizeof(double));

  lastacc = (char *) MALLOC(1*sizeof(char));
  lastacc[0] = '\0';
  lines = (List_T) NULL;
  while ((line = Getline_wlinefeed(&linelen,fp)) != NULL) {
    acc = get_acc(line);
    if (!strcmp(acc,lastacc)) {
      /* Same acc as before */
      FREE(acc);
      acc = lastacc;
    } else if (lines == NULL) {
      /* First acc seen */
      FREE(lastacc);
      lastacc = acc;
    } else {
      /* Process lines */
      prior = collect_prior_and_umis_block(prior,lines,cell_umi_tables,
					   transcript_dict,cell_dict,umi_dict);
      lines_gc(&lines);
      FREE(lastacc);
      lastacc = acc;
    }
    lines = List_push(lines,(void *) line);
  }

  if (lines == NULL) {
    /* No lines seen */
    FREE(lastacc);
  } else {
    /* Process lines */
    prior = collect_prior_and_umis_block(prior,lines,cell_umi_tables,
					 transcript_dict,cell_dict,umi_dict);
    lines_gc(&lines);
    FREE(lastacc);
  }

  return prior;
}


static UMI_T *
correct_umis (FILE *out, UMI_T *umi_array, int n_umis, char *cell_name_ptr,
	      char *transcript_name, char *gene_name) {
  UMI_T *new_array, umi_i, umi_j;
  int n_corrected = 0, i, j;

  for (i = 0; i < n_umis; i++) {
    if ((umi_i = umi_array[i]) != NULL) {
      for (j = i+1; j < n_umis; j++) {
	if ((umi_j = umi_array[j]) != NULL) {
	  if (UMI_ndiffs(umi_i,umi_j) == 1 && umi_i->nreads > umi_j->nreads) {
	    if (out != NULL) {
	      fprintf(out,"%s %s %s %s %d %s %d\n",
		      cell_name_ptr,gene_name,transcript_name,umi_j->barcode_ptr,umi_j->nreads,
		      umi_i->barcode_ptr,umi_i->nreads);
	    }

	    umi_i->nreads += umi_j->nreads;
	    if (umi_j->velocity_type > umi_i->velocity_type) {
	      umi_i->velocity_type = umi_j->velocity_type;
	    }

	    UMI_free(&umi_j);
	    umi_array[j] = (UMI_T) NULL;
	    n_corrected += 1;
	  }
	}
      }
    }
  }

  if (n_corrected == 0) {
    return umi_array;

  } else {
    new_array = (UMI_T *) MALLOC((n_umis - n_corrected + 1)*sizeof(UMI_T));

    j = 0;
    for (i = 0; i < n_umis; i++) {
      if ((umi_i = umi_array[i]) != NULL) {
	new_array[j++] = umi_i;
      }
    }
    new_array[j] = (UMI_T) NULL;

    FREE(umi_array);
    return new_array;
  }
}


static void
sort_and_correct_umis (FILE *out, Table_T *cell_umi_tables, char **all_transcripts, int n_total_transcripts,
		       int *transcripti_to_genei, char **all_genes) {
  Table_T cell_umi_table;
  int transcript_index, gene_index;

  char **cells, *cell_name_ptr, *transcript_name, *gene_name;
  int ncells, celli;

  List_T umi_list;
  UMI_T *umi_array;
  int n_umis;

  for (transcript_index = 0; transcript_index < n_total_transcripts; transcript_index++) {
    transcript_name = all_transcripts[transcript_index];
    gene_index = transcripti_to_genei[transcript_index];
    gene_name = all_genes[gene_index];
    
    if ((cell_umi_table = cell_umi_tables[transcript_index]) != (Table_T) NULL) {
      ncells = Table_length(cell_umi_table);
      cells = (char **) Table_keys(cell_umi_table,NULL);
      for (celli = 0; celli < ncells; celli++) {
	cell_name_ptr = cells[celli];
	if ((umi_list = (List_T) Table_get(cell_umi_table,(void *) cell_name_ptr)) != NULL) {
	  n_umis = List_length(umi_list);
	  umi_array = (UMI_T *) List_to_array(umi_list,NULL); /* Creates NULL at the end of the array  */
	  qsort(umi_array,n_umis,sizeof(UMI_T),UMI_cmp);
	  List_free(&umi_list);

	  if (bulk_rna_p == false) {
	    /* Correct UMIs */
	    umi_array = correct_umis(out,umi_array,n_umis,cell_name_ptr,transcript_name,gene_name);
	  }

	  /* Replace list with array */
	  Table_put(cell_umi_table,(void *) cell_name_ptr,(void *) umi_array);
	}
      }

      FREE(cells);
    }
  }

  return;
}


typedef struct Assoc_T *Assoc_T;
struct Assoc_T {
  char *barcode_ptr;
  int transcript_index;
  double likelihood;
  int nreads;			/* For future use (molecule info) */
  Velocity_T velocity_type;
};


static Assoc_T
Assoc_new (char *barcode_ptr) {
  Assoc_T new = (Assoc_T) MALLOC(sizeof(*new));

  new->barcode_ptr = barcode_ptr;
  new->transcript_index = -1;
  new->likelihood = 0.0;
  new->nreads = 0;
  new->velocity_type = UNSEEN;

  return new;
}

static void
Assoc_free (Assoc_T *old) {
  FREE(*old);
  return;
}



/* Returns cell_counts */
static Tableint_T
compute_counts (Tableint_T *transcripts_unspliced, Tableint_T *transcripts_spliced,
		Tableint_T *transcripts_ambiguous,
		Tableint_T *genes_unspliced, Tableint_T *genes_spliced, Tableint_T *genes_ambiguous,
		Table_T *cell_umi_tables, double *prior, Intlist_T *genei_to_transcripti,
		int n_total_genes, int ncells_hint) {

  Tableint_T cell_counts, transcript_count_table, gene_count_table;
  Velocity_T gene_velocity_type;
  int count;

  Tableint_T cell_expressing;
  int gene_index, transcript_index;
  Intlist_T transcript_list, p;
  double prior_value, likelihood;

  Table_T cell_umi_table, assoc_table;
  char **cells, *cell_name_ptr;
  UMI_T **umi_arrays, *ptr, umi;
  Assoc_T *assocs, assoc;
  int ncells, nassocs, celli, associ, n, j;


  cell_counts = Tableint_new(ncells_hint,NULL,NULL);

  for (gene_index = 0; gene_index < n_total_genes; gene_index++) {
    /* For each gene, gather its cells */
    cell_expressing = Tableint_new(/*hint*/1000,NULL,NULL);

    transcript_list = genei_to_transcripti[gene_index];
    for (p = transcript_list; p != NULL; p = Intlist_next(p)) {
      transcript_index = Intlist_head(p);
      if ((cell_umi_table = cell_umi_tables[transcript_index]) != NULL) {
	ncells = Table_length(cell_umi_table);
	cells = (char **) Table_keys(cell_umi_table,NULL);
	for (celli = 0; celli < ncells; celli++) {
	  cell_name_ptr = cells[celli];
	  Tableint_put(cell_expressing,(void *) cell_name_ptr,(int) true);
	}
	FREE(cells);
      }
    }

    if ((ncells = Tableint_length(cell_expressing)) > 0) {
      cells = (char **) Tableint_keys(cell_expressing,NULL);

      /* For each cell, associate UMIs with its most prevalent transcript */
      for (celli = 0; celli < ncells; celli++) {
	cell_name_ptr = cells[celli];
	
	assoc_table = Table_new(/*hint*/100000,NULL,NULL);
	
	/* printf("Processing %s\n",Intlist_to_string(transcript_list)); */
	for (p = transcript_list; p != NULL; p = Intlist_next(p)) {
	  transcript_index = Intlist_head(p);
	  prior_value = prior[transcript_index];
	  
	  if ((cell_umi_table = cell_umi_tables[transcript_index]) != NULL) {
	    if ((ptr = (UMI_T *) Table_get(cell_umi_table,(void *) cell_name_ptr)) != NULL) {
	      while ((umi = *ptr++) != NULL) {
		if ((assoc = (Assoc_T) Table_get(assoc_table,(void *) umi->barcode_ptr)) == NULL) {
		  assoc = Assoc_new(umi->barcode_ptr);
		  Table_put(assoc_table,(void *) umi->barcode_ptr,(void *) assoc);
		}
		debug(printf("Comparing likelihood %f (prior %f, nreads %d) with %f\n",
			     prior_value * umi->nreads,prior_value,umi->nreads,assoc->likelihood));
		if ((likelihood = prior_value * umi->nreads) > assoc->likelihood) {
		  assoc->likelihood = likelihood;
		  assoc->transcript_index = transcript_index;
		  assoc->nreads = umi->nreads;
		  assoc->velocity_type = umi->velocity_type;
		  debug(printf("assoc transcript_index %d\n",transcript_index));
		}
	      }
	    }
	  }
	}

	/* Store results */
	if ((nassocs = Table_length(assoc_table)) > 0) {
	  assocs = (Assoc_T *) Table_values(assoc_table,NULL);
	  gene_velocity_type = UNSEEN;
	  for (associ = 0; associ < nassocs; associ++) {
	    /* Transcript-level counting */
	    assoc = assocs[associ];
	    transcript_index = assoc->transcript_index;

	    if (assoc->velocity_type == UNSPLICED) {
	      if ((transcript_count_table = transcripts_unspliced[transcript_index]) == NULL) {
		transcript_count_table = transcripts_unspliced[transcript_index] =
		  Tableint_new(ncells_hint,NULL,NULL);
	      }
	      if (gene_velocity_type < UNSPLICED) {
		gene_velocity_type = UNSPLICED;
	      }
	      
	    } else if (assoc->velocity_type == SPLICED) {
	      if ((transcript_count_table = transcripts_spliced[transcript_index]) == NULL) {
		transcript_count_table = transcripts_spliced[transcript_index] =
		  Tableint_new(ncells_hint,NULL,NULL);
	      }
	      if (gene_velocity_type < SPLICED) {
		gene_velocity_type = SPLICED;
	      }
	      
	    } else if (assoc->velocity_type == AMBIGUOUS) {
	      if ((transcript_count_table = transcripts_ambiguous[transcript_index]) == NULL) {
		transcript_count_table = transcripts_ambiguous[transcript_index] =
		  Tableint_new(ncells_hint,NULL,NULL);
	      }
	      if (gene_velocity_type < AMBIGUOUS) {
		gene_velocity_type = AMBIGUOUS;
	      }
	      
	    } else {
	      abort();
	    }

	    count = Tableint_get(transcript_count_table,(void *) cell_name_ptr);
	    if (bulk_rna_p == true) {
	      Tableint_put(transcript_count_table,(void *) cell_name_ptr,count + assoc->nreads);
	    } else {
	      Tableint_put(transcript_count_table,(void *) cell_name_ptr,count + 1);
	    }
	    
	    /* Gene-level counting */
	    if (gene_velocity_type == UNSPLICED) {
	      if ((gene_count_table = genes_unspliced[gene_index]) == NULL) {
		gene_count_table = genes_unspliced[gene_index] =
		  Tableint_new(ncells_hint,NULL,NULL);
	      }
	    } else if (gene_velocity_type == SPLICED) {
	      if ((gene_count_table = genes_spliced[gene_index]) == NULL) {
		gene_count_table = genes_spliced[gene_index] =
		  Tableint_new(ncells_hint,NULL,NULL);
	      }
	    } else if (gene_velocity_type == AMBIGUOUS) {
	      if ((gene_count_table = genes_ambiguous[gene_index]) == NULL) {
		gene_count_table = genes_ambiguous[gene_index] =
		  Tableint_new(ncells_hint,NULL,NULL);
	      }
	    }
	    
	    count = Tableint_get(gene_count_table,(void *) cell_name_ptr);
	    if (bulk_rna_p == true) {
	      Tableint_put(gene_count_table,(void *) cell_name_ptr,count + assoc->nreads);
	    } else {
	      Tableint_put(gene_count_table,(void *) cell_name_ptr,count + 1);
	    }

	    count = Tableint_get(cell_counts,(void *) cell_name_ptr);
	    Tableint_put(cell_counts,(void *) cell_name_ptr,count + 1);

	    Assoc_free(&assoc);
	  }

	  FREE(assocs);
	}

	Table_free(&assoc_table);
      }

      FREE(cells);
    }

    Tableint_free(&cell_expressing);

    /* Free memory for all transcripts for this gene */
    for (p = transcript_list; p != NULL; p = Intlist_next(p)) {
      transcript_index = Intlist_head(p);
      if ((cell_umi_table = cell_umi_tables[transcript_index]) != NULL) {
	n = Table_length(cell_umi_table);
	umi_arrays = (UMI_T **) Table_values(cell_umi_table,NULL);

	for (j = 0; j < n; j++) {
	  ptr = umi_arrays[j];
	  while ((umi = *ptr++) != NULL) {
	    UMI_free(&umi);
	  }
	  FREE(umi_arrays[j]);
	}
	FREE(umi_arrays);
	Table_free(&cell_umi_table);
      }
    }

  }

  return cell_counts;
}




static char **
filter_cells (int *n_filtered_cells, Tableint_T cell_counts, int min_umis) {
  char **filtered_cells;
  List_T cell_list = NULL;
  char **cells;
  int *counts;
  int n, i;

  n = Tableint_length(cell_counts);
  cells = (char **) Tableint_keys(cell_counts,/*end*/NULL);
  counts = Tableint_values(cell_counts,/*end*/-1);

  for (i = 0; i < n; i++) {
    if (counts[i] >= min_umis) {
      cell_list = List_push(cell_list,(void *) cells[i]);
    }
  }
  FREE(counts);
  FREE(cells);

  filtered_cells = (char **) List_to_array_n(&(*n_filtered_cells),cell_list);
  List_free(&cell_list);

  return filtered_cells;
}


static void
print_priors (FILE *out, double *prior, char **all_transcripts, char **all_genes,
	      int *transcripti_to_genei, int n_total_transcripts) {
  int i;
  int gene_index;
  
  for (i = 0; i < n_total_transcripts; i++) {
    gene_index = transcripti_to_genei[i]; /* 0-based */
    fprintf(out,"%s\t%s\t%.2f\n",all_transcripts[i],all_genes[gene_index],prior[i]);
  }
  
  return;
}


static void
print_transcripts (FILE *out, char **all_transcripts, char **all_genes,
		   int *transcripti_to_genei, int n_total_transcripts) {
  int i;
  int gene_index;
  
  for (i = 0; i < n_total_transcripts; i++) {
    gene_index = transcripti_to_genei[i]; /* 0-based */
    fprintf(out,"%s\t%s\n",all_transcripts[i],all_genes[gene_index]);
  }
  
  return;
}


static void
print_genes (FILE *out, char **all_genes, int n_total_genes) {
  int i;
  
  for (i = 0; i < n_total_genes; i++) {
    fprintf(out,"%s\n",all_genes[i]);
  }

  return;
}


static void
print_cells (FILE *out, char **all_cells, int n_total_cells) {
  int j;
  
  for (j = 0; j < n_total_cells; j++) {
    fprintf(out,"%s\n",all_cells[j]);
  }
  
  return;
}


/* Does the well number sort first? */
static int
cell_cmp (const void *x, const void *y) {
  char *a = * (char **) x;
  char *b = * (char **) y;

  return strcmp(a,b);
}


/* Prints in MatrixMarket format */
static void
print_matrix (FILE *out, Tableint_T *count_tables, int n_total,
	      char **filtered_cells, int n_filtered_cells) {
  int nentries, n_observed_cells, i, j, k;
  char **observed_cells, *cell_name_ptr;
  Tableint_T count_table;
  int count;
  int cmp;
  
  /* Print header */
  fprintf(out,"%%%%MatrixMarket matrix coordinate integer general\n");
  fprintf(out,"%%\n");

  /* Count entries */
  nentries = 0;
  for (i = 0; i < n_total; i++) {
    if ((count_table = count_tables[i]) != NULL &&
	(n_observed_cells = Tableint_length(count_table)) > 0) {
      observed_cells = (char **) Tableint_keys(count_table,NULL);
      qsort(observed_cells,n_observed_cells,sizeof(char *),cell_cmp);

      j = k = 0;
      while (j < n_observed_cells && k < n_filtered_cells) {
	if ((cmp = cell_cmp((const void *) &(observed_cells[j]),(const void *) &(filtered_cells[k]))) < 0) {
	  j++;
	} else if (cmp > 0) {
	  k++;
	} else {
	  nentries += 1;
	  j++; k++;
	}
      }
      FREE(observed_cells);
    }
  }

  /* Print totals */
  if (matrix_order == CELLS_AS_COLUMNS) {
    fprintf(out,"%d %d %d\n",n_total,n_filtered_cells,nentries);
  } else {
    fprintf(out,"%d %d %d\n",n_filtered_cells,n_total,nentries);
  }
  
  /* Print entries */
  for (i = 0; i < n_total; i++) {
    if ((count_table = count_tables[i]) != NULL &&
	(n_observed_cells = Tableint_length(count_table)) > 0) {
      observed_cells = (char **) Tableint_keys(count_table,NULL);
      qsort(observed_cells,n_observed_cells,sizeof(char *),cell_cmp);

      j = k = 0;
      while (j < n_observed_cells && k < n_filtered_cells) {
	if ((cmp = cell_cmp((const void *) &(observed_cells[j]),(const void *) &(filtered_cells[k]))) < 0) {
	  j++;
	} else if (cmp > 0) {
	  k++;
	} else {
	  cell_name_ptr = filtered_cells[k];
	  count = Tableint_get(count_table,(void *) cell_name_ptr);
	  if (matrix_order == CELLS_AS_COLUMNS) {
	    fprintf(out,"%d %d %d\n",i+1,k+1,count);
	  } else {
	    fprintf(out,"%d %d %d\n",k+1,i+1,count);
	  }
	  j++; k++;
	}
      }
      FREE(observed_cells);
    }
  }
      
  return;
}


static void
print_single_cell (Tableint_T *transcripts_unspliced, Tableint_T *transcripts_spliced, Tableint_T *transcripts_ambiguous,
		   Tableint_T *genes_unspliced, Tableint_T *genes_spliced, Tableint_T *genes_ambiguous,
		   char **all_transcripts, int *transcripti_to_genei, int n_total_transcripts,
		   char **all_genes, int n_total_genes, char **filtered_cells, int n_filtered_cells) {
  FILE *out;
  char *transcript_file, *gene_file, *cell_file,
    *transcripts_unspliced_file, *transcripts_spliced_file, *transcripts_ambiguous_file,
    *genes_unspliced_file, *genes_spliced_file, *genes_ambiguous_file;
  int string_length;
  
  if (output_prefix == NULL) {
    transcript_file = "transcripts.tsv";
    gene_file = "genes.tsv";
    cell_file = "barcodes.tsv";
    transcripts_unspliced_file = "transcripts.unspliced.mtx";
    transcripts_spliced_file = "transcripts.spliced.mtx";
    transcripts_ambiguous_file = "transcripts.ambiguous.mtx";
    genes_unspliced_file = "genes.unspliced.mtx";
    genes_spliced_file = "genes.spliced.mtx";
    genes_ambiguous_file = "genes.ambiguous.mtx";

  } else {
    string_length = strlen(output_prefix);
    if (output_prefix[string_length - 1] == '/') {
      transcript_file = (char *) MALLOC((string_length+strlen("transcripts.tsv")+1)*sizeof(char));
      gene_file = (char *) MALLOC((string_length+strlen("genes.tsv")+1)*sizeof(char));
      cell_file = (char *) MALLOC((string_length+strlen("barcodes.tsv")+1)*sizeof(char));
      transcripts_unspliced_file = (char *) MALLOC((string_length+strlen("transcripts.unspliced.mtx")+1)*sizeof(char));
      transcripts_spliced_file = (char *) MALLOC((string_length+strlen("transcripts.spliced.mtx")+1)*sizeof(char));
      transcripts_ambiguous_file = (char *) MALLOC((string_length+strlen("transcripts.ambiguous.mtx")+1)*sizeof(char));
      genes_unspliced_file = (char *) MALLOC((string_length+strlen("genes.unspliced.mtx")+1)*sizeof(char));
      genes_spliced_file = (char *) MALLOC((string_length+strlen("genes.spliced.mtx")+1)*sizeof(char));
      genes_ambiguous_file = (char *) MALLOC((string_length+strlen("genes.ambiguous.mtx")+1)*sizeof(char));

      sprintf(transcript_file,"%stranscripts.tsv",output_prefix);
      sprintf(gene_file,"%sgenes.tsv",output_prefix);
      sprintf(cell_file,"%sbarcodes.tsv",output_prefix);
      sprintf(transcripts_unspliced_file,"%stranscripts.unspliced.mtx",output_prefix);
      sprintf(transcripts_spliced_file,"%stranscripts.spliced.mtx",output_prefix);
      sprintf(transcripts_ambiguous_file,"%stranscripts.ambiguous.mtx",output_prefix);
      sprintf(genes_unspliced_file,"%sgenes.unspliced.mtx",output_prefix);
      sprintf(genes_spliced_file,"%sgenes.spliced.mtx",output_prefix);
      sprintf(genes_ambiguous_file,"%sgenes.ambiguous.mtx",output_prefix);

    } else {
      transcript_file = (char *) MALLOC((string_length+strlen(".transcripts.tsv")+1)*sizeof(char));
      gene_file = (char *) MALLOC((string_length+strlen(".genes.tsv")+1)*sizeof(char));
      cell_file = (char *) MALLOC((string_length+strlen(".barcodes.tsv")+1)*sizeof(char));
      transcripts_unspliced_file = (char *) MALLOC((string_length+strlen(".transcripts.unspliced.mtx")+1)*sizeof(char));
      transcripts_spliced_file = (char *) MALLOC((string_length+strlen(".transcripts.spliced.mtx")+1)*sizeof(char));
      transcripts_ambiguous_file = (char *) MALLOC((string_length+strlen(".transcripts.ambiguous.mtx")+1)*sizeof(char));
      genes_unspliced_file = (char *) MALLOC((string_length+strlen(".genes.unspliced.mtx")+1)*sizeof(char));
      genes_spliced_file = (char *) MALLOC((string_length+strlen(".genes.spliced.mtx")+1)*sizeof(char));
      genes_ambiguous_file = (char *) MALLOC((string_length+strlen(".genes.ambiguous.mtx")+1)*sizeof(char));

      sprintf(transcript_file,"%s.transcripts.tsv",output_prefix);
      sprintf(gene_file,"%s.genes.tsv",output_prefix);
      sprintf(cell_file,"%s.barcodes.tsv",output_prefix);
      sprintf(transcripts_unspliced_file,"%s.transcripts.unspliced.mtx",output_prefix);
      sprintf(transcripts_spliced_file,"%s.transcripts.spliced.mtx",output_prefix);
      sprintf(transcripts_ambiguous_file,"%s.transcripts.ambiguous.mtx",output_prefix);
      sprintf(genes_unspliced_file,"%s.genes.unspliced.mtx",output_prefix);
      sprintf(genes_spliced_file,"%s.genes.spliced.mtx",output_prefix);
      sprintf(genes_ambiguous_file,"%s.genes.ambiguous.mtx",output_prefix);
    }
  }

  if ((out = fopen(transcript_file,"w")) == NULL) {
    fprintf(stderr,"Unable to write to file %s\n",transcript_file);
    exit(9);
  } else {
    print_transcripts(out,all_transcripts,all_genes,transcripti_to_genei,n_total_transcripts);
    fclose(out);
    if (output_prefix != NULL) {
      FREE(transcript_file);
    }
  }

  if ((out = fopen(gene_file,"w")) == NULL) {
    fprintf(stderr,"Unable to write to file %s\n",gene_file);
    exit(9);
  } else {
    print_genes(out,all_genes,n_total_genes);
    fclose(out);
    if (output_prefix != NULL) {
      FREE(gene_file);
    }
  }

  if ((out = fopen(cell_file,"w")) == NULL) {
    fprintf(stderr,"Unable to write to file %s\n",cell_file);
    exit(9);
  } else {
    print_cells(out,filtered_cells,n_filtered_cells);
    fclose(out);
    if (output_prefix != NULL) {
      FREE(cell_file);
    }
  }

  if ((out = fopen(transcripts_unspliced_file,"w")) == NULL) {
    fprintf(stderr,"Unable to write to file %s\n",transcripts_unspliced_file);
    exit(9);
  } else {
    print_matrix(out,transcripts_unspliced,n_total_transcripts,filtered_cells,n_filtered_cells);
    fclose(out);
    if (output_prefix != NULL) {
      FREE(transcripts_unspliced_file);
    }
  }

  if ((out = fopen(transcripts_spliced_file,"w")) == NULL) {
    fprintf(stderr,"Unable to write to file %s\n",transcripts_spliced_file);
    exit(9);
  } else {
    print_matrix(out,transcripts_spliced,n_total_transcripts,filtered_cells,n_filtered_cells);
    fclose(out);
    if (output_prefix != NULL) {
      FREE(transcripts_spliced_file);
    }
  }

  if ((out = fopen(transcripts_ambiguous_file,"w")) == NULL) {
    fprintf(stderr,"Unable to write to file %s\n",transcripts_ambiguous_file);
    exit(9);
  } else {
    print_matrix(out,transcripts_ambiguous,n_total_transcripts,filtered_cells,n_filtered_cells);
    fclose(out);
    if (output_prefix != NULL) {
      FREE(transcripts_ambiguous_file);
    }
  }


  if ((out = fopen(genes_unspliced_file,"w")) == NULL) {
    fprintf(stderr,"Unable to write to file %s\n",genes_unspliced_file);
    exit(9);
  } else {
    print_matrix(out,genes_unspliced,n_total_genes,filtered_cells,n_filtered_cells);
    fclose(out);
    if (output_prefix != NULL) {
      FREE(genes_unspliced_file);
    }
  }

  if ((out = fopen(genes_spliced_file,"w")) == NULL) {
    fprintf(stderr,"Unable to write to file %s\n",genes_spliced_file);
    exit(9);
  } else {
    print_matrix(out,genes_spliced,n_total_genes,filtered_cells,n_filtered_cells);
    fclose(out);
    if (output_prefix != NULL) {
      FREE(genes_spliced_file);
    }
  }

  if ((out = fopen(genes_ambiguous_file,"w")) == NULL) {
    fprintf(stderr,"Unable to write to file %s\n",genes_ambiguous_file);
    exit(9);
  } else {
    print_matrix(out,genes_ambiguous,n_total_genes,filtered_cells,n_filtered_cells);
    fclose(out);
    if (output_prefix != NULL) {
      FREE(genes_ambiguous_file);
    }
  }
  
  return;
}



static char *
check_valid_int (char *string) {
  char *p = string;

  if (*p == '+' || *p == '-') {
    p++;
  }

  if (!isdigit(*p)) {
    fprintf(stderr,"value %s is not a valid int\n",string);
    exit(9);
    return NULL;
  }
  while (*p != '\0' && isdigit(*p)) {
    p++;
  }

  if (*p == 'e') {
    p++;
    if (*p == '+') {
      p++;
    }
    if (!isdigit(*p)) {
      return false;
    }
    while (*p != '\0' && isdigit(*p)) {
      p++;
    }
  }

  if (*p == '\0') {
    return string;
  } else {
    fprintf(stderr,"value %s is not a valid int\n",string);
    exit(9);
    return NULL;
  }
}


static int
parse_command_line (int argc, char *argv[], int optind) {
  int opt, c;
  extern char *optarg;
  int long_option_index = 0;
  const char *long_name;
  char **argstart;
  /* char *string; */

  fprintf(stderr,"velocity-counts version %s called with args:",PACKAGE_VERSION);
  argstart = &(argv[-optind]);
  for (c = 1; c < argc + optind; c++) {
      fprintf(stderr," %s",argstart[c]);
  }
  fprintf(stderr,"\n");

  while ((opt = getopt_long(argc,argv,"D:d:C:c:o:",
			    long_options, &long_option_index)) != -1) {
    switch (opt) {
    case 0:
      long_name = long_options[long_option_index].name;
      if (!strcmp(long_name,"version")) {
	print_program_version();
	return 1;

      } else if (!strcmp(long_name,"help")) {
	print_program_usage();
	return 1;

      } else if (!strcmp(long_name,"bulk")) {
	bulk_rna_p = true;

      } else if (!strcmp(long_name,"paired-end_splicedist")) {
	paired_end_splicedist = atoi(check_valid_int(optarg));

      } else if (!strcmp(long_name,"matrix-columns")) {
	if (!strcmp(optarg,"cells")) {
	  matrix_order = CELLS_AS_COLUMNS;
	} else if (!strcmp(optarg,"genes")) {
	  matrix_order = CELLS_AS_ROWS;
	} else {
	  fprintf(stderr,"Don't recognize %s for --matrix-columns.  Allowed values: cells, genes\n",
		  optarg);
	  return 9;
	}

      } else if (!strcmp(long_name,"priors")) {
	prior_file = optarg;

      } else if (!strcmp(long_name,"umi-corrections")) {
	correction_file = optarg;

      } else if (!strcmp(long_name,"min-umis")) {
	min_umis = atoi(check_valid_int(optarg));

      } else {
	/* Shouldn't reach here */
	fprintf(stderr,"Don't recognize option %s.  For usage, run 'gsnap --help'",long_name);
	return 9;
      }
      break;

    case 'D': user_genomedir = optarg; break;
    case 'd': genome_dbroot = optarg; break;
    case 'c': transcriptome_dbroot = optarg; break;

    case 'o': output_prefix = optarg; break;

    case '?': fprintf(stderr,"For usage, run 'gsnap --help'\n"); return 9;
    default: return 9;
    }
  }

  if (genome_dbroot == NULL) {
    fprintf(stderr,"Need to specify the -d flag.  For usage, run 'velocity-counts --help'\n");
    /* print_program_usage(); */
    return 9;
  }

  if (transcriptome_dbroot == NULL) {
    fprintf(stderr,"Need to specify the -c flag.  For usage, run 'velocity-counts --help'\n");
    /* print_program_usage(); */
    return 9;
  }

  if (bulk_rna_p == true) {
    min_umis = 1;
  }

  return 0;
}


int
main (int argc, char *argv[]) {
  int cmdline_status;

  char *genomesubdir = NULL, *genome_fileroot = NULL, *dbversion;
  IIT_T transcript_map_iit;

  Table_T *cell_umi_tables;
  Tableint_T transcript_dict;	/* transcript_name -> index (0-based) */
  Table_T cell_dict, umi_dict; /* cell_name -> cell_name_ptr */
  Tableint_T cell_counts;	/* cell_name_ptr -> int */

  /* Expecting GSNAP output from transcriptome-guided genomic alignment */
  char *alignment_file, *iitfile;
  FILE *fp, *out;

  double *prior;
  int ncells_hint;
  Tableint_T *transcripts_unspliced, *transcripts_spliced, *transcripts_ambiguous, transcript_count_table;
  Tableint_T *genes_unspliced, *genes_spliced, *genes_ambiguous, gene_count_table;

  char **all_transcripts, **all_genes;
  int *transcripti_to_genei;
  Intlist_T *genei_to_transcripti, transcript_list;
  int n_total_transcripts, n_total_genes, i;
  char **filtered_cells, **all_cells, **all_umis;
  int n_filtered_cells, n, j;


  cmdline_status = parse_command_line(argc,argv,optind);
  argc -= optind;
  argv += optind;

  if (cmdline_status == 0) {
    /* okay to continue */
  } else if (cmdline_status == 1) {
    exit(0);
  } else {
    exit(cmdline_status);
  }

  ncells_hint = 10000;


  /* Read transcripts and symbols */
  genomesubdir = Datadir_find_genomesubdir(&genome_fileroot,&dbversion,user_genomedir,genome_dbroot);
  FREE(dbversion);

  /* transcript_map_iit relates transcripts to genomic coordinates and is an IIT_T */
  /* Use transcript_map_iit because it has gene symbols */
  iitfile = (char *) CALLOC(strlen(genomesubdir)+strlen("/")+
			    strlen(genome_fileroot)+strlen(".transcripts/")+
			    strlen(transcriptome_dbroot)+strlen(".align.iit")+1,sizeof(char));
  sprintf(iitfile,"%s/%s.transcripts/%s.align.iit",genomesubdir,genome_fileroot,transcriptome_dbroot);

  if ((transcript_map_iit = IIT_read(iitfile,/*name*/transcriptome_dbroot,/*readonlyp*/true,/*divread*/READ_ALL,
				     /*divstring*/NULL,/*add_iit_p*/false)) == NULL) {
    fprintf(stderr,"Transcript map file %s not found\n",iitfile);
    exit(9);
  } else {
    transcript_dict = get_transcripts(&n_total_transcripts,transcript_map_iit);
    n_total_transcripts = Tableint_length(transcript_dict);

    all_genes = get_genes(&n_total_genes,&transcripti_to_genei,&genei_to_transcripti,
			  transcript_map_iit);
    IIT_free(&transcript_map_iit);
  }

  FREE(iitfile);
  FREE(genomesubdir);
  FREE(genome_fileroot);


  cell_umi_tables = (Table_T *) CALLOC(n_total_transcripts,sizeof(Table_T));

  /* Compute priors */
  alignment_file = argv[0];
  if (bulk_rna_p == true) {
    /* For bulk analysis, compute priors and then select best prior for each read */
    if ((fp = fopen(alignment_file,"r")) == NULL) {
      fprintf(stderr,"Could not open %s\n",alignment_file);
      exit(9);
    } else {
      prior = tally_prior(fp,transcript_dict);
      fclose(fp);
    }

    if ((fp = fopen(alignment_file,"r")) == NULL) {
      fprintf(stderr,"Could not open %s\n",alignment_file);
      exit(9);
    } else {
      bulk_cell_name_ptr = bulk_cell;
      bulk_umi_barcode_ptrs[UNSEEN] = "Unseen";
      bulk_umi_barcode_ptrs[AMBIGUOUS] = "Ambiguous";
      bulk_umi_barcode_ptrs[SPLICED] = "Spliced";
      bulk_umi_barcode_ptrs[UNSPLICED] = "Unspliced";
      
      cell_dict = (Table_T) NULL;
      umi_dict = (Table_T) NULL;

      collect_umis(fp,prior,cell_umi_tables,transcript_dict);
      fclose(fp);
    }

  } else {
    /* For single-cell analysis, compute priors and collect UMIs at the same time */
    if ((fp = fopen(alignment_file,"r")) == NULL) {
      fprintf(stderr,"Could not open %s\n",alignment_file);
      exit(9);
    } else {
      cell_dict = Table_new(ncells_hint,Table_string_compare,Table_string_hash);
      umi_dict = Table_new(/*hint*/1000000,Table_string_compare,Table_string_hash);

      prior = collect_prior_and_umis(fp,cell_umi_tables,transcript_dict,cell_dict,umi_dict);
      fclose(fp);
    }
  }
    
  all_transcripts = (char **) Tableint_keys_by_timeindex(transcript_dict,NULL);

  if (prior_file == NULL) {
    /* Skip printing */
  } else if ((out = fopen(prior_file,"w")) == NULL) {
    fprintf(stderr,"Could not write to %s\n",prior_file);
    exit(9);
  } else {
    print_priors(out,prior,all_transcripts,all_genes,transcripti_to_genei,n_total_transcripts);
    fclose(out);
  }
    

  /* Sort and correct UMIs */
  if (bulk_rna_p == true) {
    out = (FILE *) NULL;
  } else if (correction_file == NULL) {
    out = (FILE *) NULL;
  } else if ((out = fopen(correction_file,"w")) == NULL) {
    fprintf(stderr,"Could not write to %s\n",correction_file);
    exit(9);
  }

  sort_and_correct_umis(out,cell_umi_tables,all_transcripts,n_total_transcripts,
			transcripti_to_genei,all_genes);
  if (out != NULL) {
    fclose(out);
  }


  /* Compute counts */
  transcripts_unspliced = (Tableint_T *) CALLOC(n_total_transcripts,sizeof(Tableint_T));
  transcripts_spliced = (Tableint_T *) CALLOC(n_total_transcripts,sizeof(Tableint_T));
  transcripts_ambiguous = (Tableint_T *) CALLOC(n_total_transcripts,sizeof(Tableint_T));

  genes_unspliced = (Tableint_T *) CALLOC(n_total_genes,sizeof(Tableint_T));
  genes_spliced = (Tableint_T *) CALLOC(n_total_genes,sizeof(Tableint_T));
  genes_ambiguous = (Tableint_T *) CALLOC(n_total_genes,sizeof(Tableint_T));

  cell_counts = compute_counts(transcripts_unspliced,transcripts_spliced,transcripts_ambiguous,
			       genes_unspliced,genes_spliced,genes_ambiguous,cell_umi_tables,
			       prior,genei_to_transcripti,n_total_genes,ncells_hint);
  FREE(prior);


  /* Print results */
  filtered_cells = filter_cells(&n_filtered_cells,cell_counts,min_umis);
  qsort(filtered_cells,n_filtered_cells,sizeof(char *),cell_cmp);
  Tableint_free(&cell_counts);

  print_single_cell(transcripts_unspliced,transcripts_spliced,transcripts_ambiguous,
		    genes_unspliced,genes_spliced,genes_ambiguous,
		    all_transcripts,transcripti_to_genei,n_total_transcripts,
		    all_genes,n_total_genes,filtered_cells,n_filtered_cells);
  FREE(filtered_cells);


  /* Free memory */
  FREE(transcripti_to_genei);

  for (i = 0; i < n_total_genes; i++) {
    transcript_list = genei_to_transcripti[i];
    Intlist_free(&transcript_list);
    FREE(all_genes[i]);
  }
  FREE(all_genes);
  FREE(genei_to_transcripti);

  for (i = 0; i < n_total_transcripts; i++) {
    FREE(all_transcripts[i]);
  }
  FREE(all_transcripts);
  Tableint_free(&transcript_dict);

  if (bulk_rna_p == false) {
    n = Table_length(cell_dict);
    all_cells = (char **) Table_keys(cell_dict,NULL);
    for (j = 0; j < n; j++) {
      FREE(all_cells[j]);
    }
    FREE(all_cells);
    Table_free(&cell_dict);
  }

  if (bulk_rna_p == false) {
    n = Table_length(umi_dict);
    all_umis = (char **) Table_keys(umi_dict,NULL);
    for (i = 0; i < n; i++) {
      FREE(all_umis[i]);
    }
    FREE(all_umis);
    Table_free(&umi_dict);
  }

  for (i = 0; i < n_total_genes; i++) {
    /* Note: keys were pointers to entries in all_cells */
    if ((gene_count_table = genes_unspliced[i]) != NULL) {
      Tableint_free(&gene_count_table);
    }
    if ((gene_count_table = genes_spliced[i]) != NULL) {
      Tableint_free(&gene_count_table);
    }
    if ((gene_count_table = genes_ambiguous[i]) != NULL) {
      Tableint_free(&gene_count_table);
    }
  }
  FREE(genes_unspliced);
  FREE(genes_spliced);
  FREE(genes_ambiguous);

  for (i = 0; i < n_total_transcripts; i++) {
    /* Note: keys were pointers to entries in all_cells */
    if ((transcript_count_table = transcripts_unspliced[i]) != NULL) {
      Tableint_free(&transcript_count_table);
    }
    if ((transcript_count_table = transcripts_spliced[i]) != NULL) {
      Tableint_free(&transcript_count_table);
    }
    if ((transcript_count_table = transcripts_ambiguous[i]) != NULL) {
      Tableint_free(&transcript_count_table);
    }
  }
  FREE(transcripts_unspliced);
  FREE(transcripts_spliced);
  FREE(transcripts_ambiguous);

#if 0
  /* This memory was freed in compute_counts */
  for (i = 0; i < n_total_transcripts; i++) {
    /* Note: keys were pointers to entries in all_cells */
    if (cell_umi_tables[i] != NULL) {
      n = Table_length(cell_umi_tables[i]);
      umi_arrays = (UMI_T **) Table_values(cell_umi_tables[i],NULL);

      for (j = 0; j < n; j++) {
	ptr = umi_arrays[j];
	while ((umi = *ptr++) != NULL) {
	  UMI_free(&umi);
	}
	FREE(umi_arrays[j]);
      }
      FREE(umi_arrays);
      Table_free(&(cell_umi_tables[i]));
    }
  }
#endif

  FREE(cell_umi_tables);

  return 0;
}


 static void
print_program_usage () {
  fprintf(stdout,"\
Usage: velocity-counts [OPTIONS...] <GSNAP alignment file>\n\
\n\
");

  /* Input options */
  fprintf(stdout,"Options (must include -d and -c)\n");
  fprintf(stdout,"\
  -D, --dir=directory            Genome directory.  Default (as specified by --with-gmapdb to the configure program) is\n\
                                   %s\n\
",GMAPDB);
  fprintf(stdout,"\
  -d, --db=STRING                Genome database\n\
\n\
");

  fprintf(stdout,"\
  -C, --transcriptdir=directory  Transcriptome directory.  Default is the value for --dir above\n\
  -c, --transcriptdb=STRING      Transcriptome database\n\
\n\
");

  fprintf(stdout,"\
  --bulk                         Handle as bulk RNA-seq (ignoring the CB and UR/UB fields)\n\
  --input-command                Process alignment file using this command (e.g., \"samtools view\" or \"gunzip -c\")\n\
  --paired-end-splicedist=INT    Distance for treating paired-end ambigous (either spliced or unspliced) as spliced\n\
                                   (default = 1000)\n\
  -o, --output=STRING            Prefix for output files.  Can be a directory if it ends with '/'\n\
  --umi-corrections=STRING       Write UMI corrected barcodes to this file\n\
  --min-umis=INT                 Print only cells having this many gene UMIs or more (default: 1000, or 1 for --bulk)\n\
  --matrix-columns=STRING        Order for matrix files. Values: cells (cells as columns, default) or genes (cells as rows)\n\
                                   The default (cells as columns) is standard for 10X Genomics, but loom files store cells\n\
                                   as rows and require a transpose (such as do.transpose=T in the loomR::create function)\n\
\n\
");

  return;
 }
